from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *
from PymoNNto.NetworkBehaviour.Input.Images.Helper import *
import hickle as hkl
import os.path
from PymoNNto.NetworkBehaviour.Input.Images.PixelPattern import *

mnist_folder = '../../../MNIST'
mnist_preprocessed_patterns = mnist_folder+'/preprocessed_patterns.data'
mnist_preprocessed_labels = mnist_folder+'/preprocessed_labels.data'

class MNIST_Patterns(Pixel_Pattern):

    def get_MNIST_activation_data(self, pic_numbers=None):
        result = []
        labels = []

        if not hasattr(self, "mndata"):
            from mnist import MNIST  # pip install python-mnist
            self.mndata = MNIST(mnist_folder)
            data = self.mndata.load_testing()
            self.mnist_pictures = data[0]
            self.mnist_labels = data[1]

        for i, l in enumerate(self.mnist_labels):
            if pic_numbers is None or l in pic_numbers:
                act = np.array(self.mnist_pictures[i]).reshape(28, 28)#/255
                #act = self.edge_detector(self.mnist_pictures[i])#[d/255*1.4 for d in self.edge_detector(self.mnist_pictures[i])]
                result.append(act)
                labels.append(l)

        return result, labels

    def get_current_pattern_index(self):
        self.label_repeat_counter += 1

        if self.label_repeat_counter>self.repeat_same_label_time:
            self.label_repeat_counter = 1
            r = np.random.randint(len(self.patterns))
            self.label_repeat_label = self.labels[r]
            return r
        else:
            for _ in range(1000):#max trys to find same label
                r = np.random.randint(len(self.patterns))
                if self.labels[r] == self.label_repeat_label:
                    return r

        return np.random.randint(len(self.patterns))

    def initialize(self):
        super().initialize()
        self.grid_width = self.kwargs.get('grid_width', 28)
        self.grid_height = self.kwargs.get('grid_height', 28)

        self.repeat_same_label_time = self.kwargs.get('repeat_same_label_time', 1)
        self.label_repeat_counter=self.repeat_same_label_time
        self.label_repeat_label=''

        self.grid_channels = 2
        if not os.path.isfile(mnist_preprocessed_patterns) or not os.path.isfile(mnist_preprocessed_labels):
            data, labels = self.get_MNIST_activation_data()
            for d, l in zip(data, labels):
                onc, offc = get_LOG_On_Off(d)
                self.patterns.append(np.array([onc/np.max(onc), offc/np.max(offc)]))
                self.labels.append(l)
            hkl.dump(self.patterns, open(mnist_preprocessed_patterns, 'w'))
            hkl.dump(self.labels, open(mnist_preprocessed_labels, 'w'))
        else:
            self.patterns=hkl.load(open(mnist_preprocessed_patterns))
            self.labels=hkl.load(open(mnist_preprocessed_labels))

        print('MNIST loaded')


    #def get_pattern(self):
    #    i = 0
    #    submat = np.zeros(1)
    #    while i < 10 and sum(submat) < 1.0:
    #        i += 1
    #        submat = self.get_image_patch()

    #    if self.patch_norm:
    #        d = np.max(submat)
    #        if d == 0:
    #            d = 1
    #        return submat/d
    #    else:
    #        return submat


    #def get_pattern_dimension(self, remove_depth=True):
    #    if remove_depth:
    #        return [self.grid_height * self.dim_matrices.shape[0], self.grid_width]
    #    else:
    #        return [self.dim_matrices.shape[0], self.grid_height, self.grid_width]

    #def reconstruct_pattern(self, p):
    #    return p.reshape(self.grid_height * self.dim_matrices.shape[0], self.grid_width)


#test = MNIST_Patterns()

#num,lab = test.get_MNIST_activation_data([2])

#onc, offc = get_LOG_On_Off(num[0])

#offc=offc.astype(np.float64)
#offc/=np.max(offc)

#plt.matshow(offc)
#plt.show()



import subprocess
import shutil
import datetime
import time
import os
import paramiko
from scp import SCPClient

import zipfile



def zipDir(dirPath, zipPath, filter):
    zipf = zipfile.ZipFile(zipPath , mode='w')
    lenDirPath = len(dirPath)
    for root, _, files in os.walk(dirPath):
        for file in files:
            filter_file=False
            for f in filter:
                if f in root or f in file+'\\':
                    filter_file = True
            if not filter_file:
                filePath = os.path.join(root, file)
                print(filePath)
                zipf.write(filePath , filePath[lenDirPath :] )
    zipf.close()



#def zipdir(path, ziph):
#    # ziph is zipfile handle
#    for root, dirs, files in os.walk(path):
#        for file in files:
#            ziph.write(os.path.join(root, file))

class Evolution_Server_Base:

    def __init__(self, name, main_path):
        self.name = name
        self.main_path = main_path
        self.temp_zip_dir = '../../../temp_zip_evolution/'

    def exec_cmd(self, cmd, shell=False, cwd=None):
        if type(cmd) is tuple:
            if len(cmd)>2:
                cwd=cmd[2]
            shell = cmd[1]
            cmd = cmd[0]
        print(cmd)

        #output = subprocess.run(cmd, stdout=subprocess.PIPE, shell=True).stdout.decode('utf-8')
        if cmd is not None:
            output = subprocess.run(cmd, stdout=subprocess.PIPE, shell=shell, cwd=cwd).stdout.decode('utf-8')
            print(output)
            return output
        else:
            return ''

    def transfer(self, evo_name):
        None

    def execute(self, evo_name, arguments):
        self.exec_cmd(self.get_Evo_Execute_cmd(evo_name, arguments))

    def terminate_screen(self, evo_name):
        self.exec_cmd(self.get_terminate_screen_cmd(evo_name))

    def remove_evo(self, evo_name):
        self.exec_cmd(self.get_remove_cmd(evo_name))

    def parse_plot_data_output(self, output):
        y = []

        for line in output.replace('\r', '').split('\n'):
            data = line.split(' ')
            if len(data) == 2 and data[0].isdigit() and data[1].replace('.', '').replace('-', '').isdigit():
                y.append(float(data[1]))

        if 'Pfad nicht finden' in output or len(y) == 0:
            print('Keine Daten gefunden...')

        return y

    def plot_data(self, evo_name):
        output = self.exec_cmd(self.get_plot_cmd(evo_name))
        return self.parse_plot_data_output(output)


    def running_evos(self):
        result = []
        output = self.exec_cmd(self.get_running_list_cmd())

        for l in output.splitlines():
            if '	' in l:
                dot_split = l.split('.')
                if len(dot_split) > 1:
                    space_split = dot_split[1].split('	')
                    if len(space_split) > 0:
                        result.append(space_split[0])
                        #print('found' + space_split[0])

        return result

    def evolution_list(self):
        result = {}

        output = self.exec_cmd(self.get_evo_list_cmd())

        for l in output.splitlines():

            while '  ' in l:
                l = l.replace('  ', ' ')

            space_split = l.split(' ')
            if len(space_split) == 9 and l[-1] != '.':

                date_time_obj = datetime.datetime.strptime(space_split[5].replace('Mai', 'May')+' '+space_split[6]+' '+space_split[7], '%b %d %H:%M')
                result[date_time_obj] = space_split[-1]


        result_list = []
        for key in reversed(sorted(list(result.keys()))):
            result_list.append(result[key])

        return result_list

    def open_terminal(self, evo_name):
        print(self.get_open_terminal_cmd(evo_name))
        os.system("start "+" ".join(self.get_open_terminal_cmd(evo_name)))#start "C:/Windows/System32/WindowsPowerShell/v1.0/powershell.exe Get-Process"') # @cmd /k " + " ".join(self.get_open_terminal_cmd(evo_name))

    def get_open_terminal_cmd(self, evo_name):
        return 'screen -r '+evo_name

    def get_plot_cmd(self, evo_name):
        return None

    def get_Evo_Execute_cmd(self, evo_name, arguments):
        return None

    def get_transfer_cmd(self, evo_name):
        return None

    def get_extract_cmd(self, evo_name):
        return None

    def get_python_exec(self, arguments, path='', python='python3'):
        screen_cmd = python + " "+path+"Multithreaded_Evolution.py"
        for arg in arguments:
            screen_cmd += " "+arg.replace(' ', '')
        return screen_cmd

    def get_terminate_screen_cmd(self, evo_name):
        return 'screen -XS '+evo_name+' quit'

    def get_remove_cmd(self, evo_name):
        return 'cd '+self.main_path+'; rm -r '+evo_name

    def get_evo_list_cmd(self):
        return 'cd '+self.main_path+'; ls -la'

    def get_running_list_cmd(self):
        return 'screen -ls'

    def get_plot_cmd(self, evo_name):
        return 'cd ' + self.main_path + evo_name + "/Exploration/Evolution_Old/; python3 Evo_Plots.py"

    def get_Evo_Execute_cmd(self, evo_name, arguments):
        pexc=self.get_python_exec(arguments)
        commands = 'cd ' + self.main_path + evo_name + '/Exploration/Evolution_Old/; screen -dmS ' + evo_name + ' sh; screen -S ' + evo_name + ' -X stuff "' + pexc + ' \r\n"'
        return commands

class Evolution_Server_Local_Windows(Evolution_Server_Base):

    def get_terminate_screen_cmd(self, evo_name):
        print('not supported for windows')
        return None

    def get_running_list_cmd(self):
        print('not supported for windows')
        return None#'C:\\cygwin64\\bin\\screen.exe -ls'

    def get_remove_cmd(self, evo_name):
        print(self.main_path + evo_name)
        shutil.rmtree(self.main_path + evo_name)
        return None

    def evolution_list(self):
        result = {}

        output = self.exec_cmd(self.get_evo_list_cmd())

        for l in output.splitlines():

            while '  ' in l:
                l = l.replace('  ', ' ')

            space_split = l.split(' ')
            if len(space_split) == 4 and l[-1] != '.' and space_split[0]!='':
                date_time_obj = datetime.datetime.strptime(space_split[0]+' '+space_split[1], '%d.%m.%Y %H:%M')
                result[date_time_obj] = space_split[-1]


        result_list = []
        for key in reversed(sorted(list(result.keys()))):
            result_list.append(result[key])

        print(result)

        return result_list

    def __init__(self, name, main_path):
        super().__init__(name, main_path)
        self.temp_zip_dir = main_path

    def get_evo_list_cmd(self):
        return 'dir '+self.main_path.replace('/', '\\'), True #-la does not exist

    def get_extract_cmd(self, evo_name):
        shutil.unpack_archive(self.main_path + evo_name + '.zip', self.main_path + evo_name , 'zip')
        #commands = 'dir'#,'Expand-Archive ' + self.main_path.replace('/','\\') + evo_name + '.zip' #unzip=>Expand-Archive + no zip delete (same device)
        return None

    #def get_Evo_Execute_cmd(self, evo_name, arguments):
        #return commands.replace('/', '\\'), True, path.replace('/', '\\')

    def get_plot_cmd(self, evo_name):
        return 'python' + self.main_path + evo_name + '/Exploration/Evolution_Old/Evo_Plots.py'

    def execute(self, evo_name, arguments):
        path = self.main_path + evo_name + '/Exploration/Evolution_Old/'
        path=path.replace('/', '\\')
        pexc = self.get_python_exec(arguments, python='python')
        os.system("start cmd.exe "+path+" @cmd /k "+pexc)

        #self.exec_cmd('screen -d -m '+pexc, True, path)

        #c1='C:\\cygwin64\\bin\\screen.exe -dmS ' + evo_name
        #self.exec_cmd(c1.replace('/', '\\'), True, path)
        #c2='C:\\cygwin64\\bin\\screen.exe -S ' + evo_name + ' -X stuff "' + pexc + ' \r\n"'
        #self.exec_cmd(c2.replace('/', '\\'), True, path)

class Evolution_Server_SSH(Evolution_Server_Base):

    def __init__(self, name, main_path, ssh_host, ssh_name, ssh_pw):
        super().__init__(name, main_path)
        self.ssh_host = ssh_host
        self.ssh_name = ssh_name
        self.ssh_pw = ssh_pw

    def ssh_wrap_cmd(self, commands):
        return ["C:\\Windows\\System32\\OpenSSH\\ssh.exe", self.ssh_name+'@'+self.ssh_host, "-t", commands]

    def transfer(self, evo_name):
        #compress

        #shutil.make_archive(self.temp_zip_dir+evo_name, 'zip', '../../../PymoNNto_Dev')
        zipDir('../../../PymoNNto_Dev/', self.temp_zip_dir+evo_name+'.zip', ['.git', '.idea', '\\StorageManager\\', '\\NetworkStates\\', '\\Evo\\', '\\__pycache__\\', '\\midis\\'])

        #transfer
        #self.exec_cmd(self.get_transfer_cmd(evo_name))

        ssh = self.get_ssh_connection()
        scp = SCPClient(ssh.get_transport())
        src = self.temp_zip_dir+evo_name+'.zip'
        dst = self.main_path+evo_name+'.zip'
        scp.put(src, dst)

        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(self.get_extract_cmd(evo_name))
        ssh_stdout.channel.recv_exit_status()

        #extract
        #self.exec_cmd(self.get_extract_cmd(evo_name))
        os.remove(self.temp_zip_dir+evo_name+'.zip')

        ssh.close()
        scp.close()

    def get_ssh_connection(self):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.ssh_host, username=self.ssh_name, password=self.ssh_pw)
        return ssh

    def exec_cmd(self, cmd, shell=False, cwd=None):
        print(cmd)

        try:
            ssh = self.get_ssh_connection()
            ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
            ssh_stdout.channel.recv_exit_status()
            result = ssh_stdout.read().decode('utf-8')
            print(result)
            ssh.close()
        except:
            result = ''

        #if type(cmd) is tuple:
        #    if len(cmd)>2:
        #        cwd=cmd[2]
        #    shell = cmd[1]
        #    cmd = cmd[0]
        #print(cmd)

        #output = subprocess.run(cmd, stdout=subprocess.PIPE, shell=True).stdout.decode('utf-8')
        #if cmd is not None:
        #    output = subprocess.run(cmd, stdout=subprocess.PIPE, shell=shell, cwd=cwd).stdout.decode('utf-8')
        #    print(output)
        #    return output
        #else:
        #    return ''

        return result

    def get_extract_cmd(self, evo_name):
        commands = 'mkdir ' + self.main_path + evo_name + '; cd '+self.main_path+'; unzip ' +evo_name+'.zip -d '+ evo_name + '; rm '+evo_name+'.zip'
        return commands#self.ssh_wrap_cmd(commands)

    #def get_terminate_screen_cmd(self, evo_name):
    #    return self.ssh_wrap_cmd(super().get_terminate_screen_cmd(evo_name))

    #def get_remove_cmd(self, evo_name):
    #    return self.ssh_wrap_cmd(super().get_remove_cmd(evo_name))

    def get_open_terminal_cmd(self, evo_name):
        return self.ssh_wrap_cmd(super().get_open_terminal_cmd(evo_name))

    #def get_evo_list_cmd(self):
    #    return self.ssh_wrap_cmd(super().get_evo_list_cmd())

    #def get_running_list_cmd(self):
    #    return self.ssh_wrap_cmd(super().get_running_list_cmd())ssh_wrap_cmd

    #def get_plot_cmd(self, evo_name):
    #    return self.ssh_wrap_cmd(super().get_plot_cmd(evo_name))

    def get_transfer_cmd(self, evo_name):
        src = self.temp_zip_dir+evo_name+'.zip'
        dst = self.ssh_target + ':' + self.main_path
        return 'scp ' + src + ' ' + dst

    #def get_Evo_Execute_cmd(self, evo_name, arguments):
    #    return self.ssh_wrap_cmd(super().get_Evo_Execute_cmd(evo_name, arguments))

class Evolution_Server_SSH_Slurm(Evolution_Server_SSH):

    def __init__(self, name, main_path, ssh_host, ssh_name, ssh_pw, slurm_wrapper, slurm_partition):
        super().__init__(name, main_path, ssh_host, ssh_name, ssh_pw)
        self.slurm_wrapper = slurm_wrapper
        self.slurm_partition = slurm_partition

    def get_terminate_screen_cmd(self, evo_name):
        cmd='scancel --name ' + evo_name
        return cmd
        #return self.ssh_wrap_cmd()

    def terminate_screen(self, evo_name):
        self.exec_cmd(self.get_terminate_screen_cmd(evo_name))
        time.sleep(2)

    def get_running_list_cmd(self):
        #cmd='sacct --format="JobID,JobName%30"'
        cmd='squeue --format="%.15i %.25j %.8u %.10M %.2t %.9P" -u vieth -p '+self.slurm_partition
        return cmd
        #return self.ssh_wrap_cmd()

    def get_slurm_Evo_Execute_cmd(self,job_name, pexc, command='sbatch'):
        return self.slurm_wrapper.replace('*command*', command) + ' --job-name='+ job_name + pexc#' --wrap="' + pexc + '"'

    def get_Evo_Execute_cmd(self, evo_name, arguments):
        pexc=self.get_python_exec(arguments)
        commands = 'cd ' + self.main_path + evo_name + '/Exploration/Evolution_Old/; ' + self.get_slurm_Evo_Execute_cmd(evo_name, pexc, command='sbatch')
        return commands
        #return self.ssh_wrap_cmd(commands)

    def running_evos(self):
        result = []
        output = self.exec_cmd(self.get_running_list_cmd())

        for l in output.splitlines():

            while '  ' in l:
                l = l.replace('  ', ' ')

            if not 'NAME' in l:
                dot_split = l.split(' ')
                if len(dot_split) > 2:
                    result.append(dot_split[2])

        print(result)

        return result




import random

import sys
import os
import threading

from PymoNNto.Exploration.StorageManager.StorageManager import *


def string_to_array(s):
    s=s.replace(' ','')
    result = []
    s = s.replace(']', '')
    parts = s.split('[')
    for p in parts:
        if p != '':
            result.append([float(num) for num in p.split(',') if num != ''])
    return result

class Individual:
    def __init__(self, params, lifetime=0):
        self.parameters = params
        self.fitness = -1
        self.lifetime = lifetime



class Evolution:
    def __init__(self, evaluation_function, max_individual_count, generations=None, name='Evolution_Old', mutation = 0.1, constraints=[], death_rate=0.5, param={}):
        self.datafolder=get_data_folder()+'/Evo/'

        if not os.path.exists(self.datafolder):
            try:
                os.mkdir(self.datafolder)
            except:
                print('evolution folder already exits')

        self.name = name#+'_{}'.format(int(np.random.rand()*10000))
        #print(self.name)
        self.param=param
        self.savestate = True
        self.evaluation_function = evaluation_function
        self.max_individual_count = max_individual_count
        self.generations = generations
        self.individuals = []
        self.mutation = mutation
        self.constraints = constraints
        self.death_rate = death_rate
        self.plot_progress=False
        self.paused=False


    def breed(self):

        new_ind = []

        weights = np.array(self.set_get_ind_params("lifetime")).copy()
        #weights -= min(weights)
        weights = weights / np.sum(weights)

        for _ in range(self.max_individual_count-len(self.individuals)):

            parent1 = np.random.choice(self.individuals, 1, p=weights)[0]
            parent2 = np.random.choice(self.individuals, 1, p=weights)[0]

            ind = [] #new individuals
            for p_i in range(len(parent1.parameters)):
                p1 = np.clip(np.random.normal(parent1.parameters[p_i], parent1.parameters[p_i]*self.mutation), 0, None)
                p2 = np.clip(np.random.normal(parent2.parameters[p_i], parent2.parameters[p_i]*self.mutation), 0, None)
                ind.append(random.choice([p1, p2]))

            for constraint in self.constraints:
                exec(constraint)

            new_ind.append(Individual(ind))

        self.individuals += new_ind

    def natural_selection(self):

        kill_count = int(len(self.individuals)*self.death_rate)

        for _ in range(kill_count):
            kill_ind = self.individuals[0]
            for ind in self.individuals:
                if ind.fitness < kill_ind.fitness:
                    kill_ind = ind

            self.individuals.remove(kill_ind)

    def set_get_ind_params(self, attr, values=None):
        result = []
        for i, ind in enumerate(self.individuals):
            if values is not None and i < len(values):
                setattr(ind, attr, values[i])
            result.append(getattr(ind, attr))
        return result


    def get_fitnesses(self, living_individuals):
        return [self.evaluation_function(self.name, individual, self.param) for individual in living_individuals]



    def start(self, seed_individuals):
        #self.start_input_thread()
        self.individuals.clear()

        generation = 0

        self.save('Seed:', seed_individuals)
        for ind in seed_individuals:

            for constraint in self.constraints:
                exec(constraint)

            self.individuals.append(Individual(ind, 1))

        self.breed()
        self.save('Start:', self.set_get_ind_params('parameters'))

        while self.generations is None or generation < self.generations:

            while self.paused:
                print('paused...')
                time.sleep(10)

            generation += 1

            fitnesses = self.get_fitnesses(self.set_get_ind_params('parameters'))
            self.set_get_ind_params('fitness', fitnesses)

            #if self.plot_progress:
            #    plt.plot(np.zeros(len(fitnesses))+generation, fitnesses)
            #    plt.show()

            new_lt = (np.array(self.set_get_ind_params('lifetime'))+np.array(self.set_get_ind_params('fitness', fitnesses)))/2
            self.set_get_ind_params('lifetime', new_lt)


            self.natural_selection()
            self.save('Gen {} survivers:'.format(generation), self.set_get_ind_params('parameters'))
            self.save('Gen {} fitnesses:'.format(generation), self.set_get_ind_params('fitness'))
            self.save('Gen {} lifetime:'.format(generation), self.set_get_ind_params('lifetime'))
            self.breed()
            self.save('Gen {} breed:'.format(generation), self.set_get_ind_params('parameters'))


    def get_evo_file(self):
        return self.datafolder+self.name+".txt"

    def evo_file_exists(self):
        return os.path.exists(self.get_evo_file())

    def save(self, message, data):
        print(message, data)
        if self.savestate:
            f = open(self.get_evo_file(), "a")
            f.write(message+" {}".format(data)+"\n")
            f.close()




    def continue_evo(self, file):
        f = open(self.datafolder+file, "r")
        population = None
        for line in f.readlines():
            if 'survivers:' in line:
                population=string_to_array(line[line.index('['):-1])
                print(population)
        f.close()

        if population is not None:
            name=file.split('.')[0]
            print('setting target file to '+name)
            self.name = name
            print('starting evolution...')
            self.save('Continue...', population)
            self.start(population)
        else:
            print('no individuals found')

    #add history saving
    #plot progress in UI Window
    #add plot and dim reduction

    def start_input_thread(self):

        def input_collect_function():
            while True:
                for line in sys.stdin:
                    if line and line != '\n':
                        line=line.replace('\r', '').replace('\n', '')
                        if '[' in line and ']' in line:
                            print('adding individual(s)...')
                            genes = string_to_array(line)
                            for gene in genes:
                                if len(self.individuals[0].parameters) == len(gene):
                                    ind = Individual(genes)
                                    ind.fitness = np.max(self.set_get_ind_params('fitness'))
                                    ind.lifetime = np.max(self.set_get_ind_params('lifetime'))
                                    self.individuals.append(ind)
                                    print('Individial added:', gene)
                                else:
                                    print('wrong number of genes in', gene)
                        elif 'pause' in line:
                            self.paused = True
                            print('evolution stopped')
                        elif 'start' in line:
                            self.paused = False
                            print('evolution continued')
                        elif line.isdigit():
                            if hasattr(self,'thread_count'):
                                self.thread_count=int(line)
                                print('thread count set to ' + line)
                        elif line.replace('.','').isdigit():
                            self.mutation=float(line)
                            print('mutation rate set to', line)
                        else:
                            print('unknown command' + line)
                time.sleep(1)

        threading.Thread(target=input_collect_function).start()



from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.sidebar_grammar_module import *
from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.sidebar_image_module import *
from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.reconstruction_tab import *
from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.sidebar_music_module import *
from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.sidebar_drumbeat_module import *
from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.sun_gravity_plot_tab import *
from PymoNNto.Exploration.Network_UI.Sequence_Activation_Tabs.character_activation_tab import *

def get_my_default_UI_modules():
    return [
        # chain_tab(),
        sun_gravity_plot_tab(),
        sidebar_image_module(),
        sidebar_grammar_module(),
        sidebar_music_module(),
        sidebar_drumbeat_module(),
        reconstruction_tab(),
        character_activation_tab()
        ]


import numpy as np



def relu2(x, param):
    return np.clip((x > param)*(1 / (1 - param) * (x - param)), 0, 1)

def relu3(x ,param, stepheight):
    return np.clip((x > param) * (stepheight + (1-stepheight) / (1 - param) * (x - param)), 0, 1)

def roll(mat):
    #return np.roll(mat, 1)
    mat[1:len(mat)]=mat[0:len(mat)-1]
    return mat


#print(GammaBlock_to_Hz(0.5))

#import matplotlib.pyplot as plt
#plt.plot([relu3(x/100, 0.1, 0) for x in range(100)])
#plt.show()


from sympy import symbols
from sympy import solve, symbols, pi, Eq
from sympy.physics.units.systems import SI
from sympy.physics.units import *
from sympy.physics.units import gravitational_constant as G
from sympy.physics.units.systems.si import dimsys_SI
from sympy.physics.units.util import *

import numpy as np

import matplotlib.pyplot as plt

#vec = np.array([1,2,3])



#x = 10*mV

#y = 100*mV

#print(float(x/y))


#F = mass*acceleration
#Dimension(acceleration*mass)
#dimsys_SI.get_dimensional_dependencies(F)
#{'length': 1, 'mass': 1, 'time': -2}
#dimsys_SI.get_dimensional_dependencies(force)
#{'length': 1, 'mass': 1, 'time': -2}


#from sympy.physics.units import Quantity
#def unit(expr):
#    return expr.subs({x: 1 for x in expr.args if not x.has(Quantity)})



#class cl:
#    def __init__(self):
#        self.test=1

#c=cl()

#print(c.__dict__)

#mV = milli*volts

#tau = 1*ms


#tau = solve(tau,T)
#c.x = symbols('c.x')
#setattr(c, 'x', symbols('c.x'))
#print(tau)

#v = (0*volts-c.x*volts)/tau

#dv/dt = (5-v)/tau
#convert_to
#u=unit(v)

#print(unit(v))
#print(convert_to(v*seconds/u, T)) #
#print(check_dimensions(v))





tau = 10*ms#*seconds


print(convert_to(tau, seconds)/second)

scale=float(convert_to(seconds/tau,seconds))

print(scale)

x = 100*volt
# = #/tau

values = []
for i in range(100):
    x = x+(0*volt-x)/scale    #v + (0*volts - v)/(tau/ms)#convert_to(v, volts)
    values.append(x / volts)
    print(x)

plt.plot(values)
plt.show()



from PymoNNto.Exploration.Evolution.Evolution_Device_Single_Thread import *
from PymoNNto.Exploration.Evolution.Evolution_Device_Multi_Thread import *
from PymoNNto.Exploration.Evolution.Evolution_Device_SSH import *
from PymoNNto.Exploration.Evolution.Evolution_Base import *
from PymoNNto.Exploration.Evolution.Breed_And_Selection_Module import *
import os.path
import time

class Evolution:

    def __init__(self, name, slave_file, individual_count=10, mutation=0.4, death_rate=0.5, devices={'single_thread':1}, constraints=[], start_genomes=[], inactive_genome_info={}, ui=True):

        self.Breed_And_Select = Default_Breed_And_Select(self, death_rate=death_rate, mutation=mutation, individual_count=individual_count, constraints=constraints)

        self.name = name

        self.slave_file = slave_file

        self.devices = []
        self.start_genomes = start_genomes
        self.inactive_genome_info = inactive_genome_info
        self.ui = ui
        self.id_counter = 0

        if not os.path.isfile(slave_file):
            print('warning slave file not found')

        #check file has set_score_function => warning

        #initialize ui

        if start_genomes==[]:
            print('Error no genomes found')

        if 'auto_detect' in start_genomes:
            print('auto genome detection')

        self.scored_individuals = []
        self.running_individuals = []

        for genome in self.start_genomes:
            if not self.Breed_And_Select.is_valid_genome(genome):
                print('Error: start genome does not fulfill constraints', genome)

        self.non_scored_individuals = self.Breed_And_Select.breed(start_genomes)
        print(self.non_scored_individuals)

        for device_string, number_of_threads in devices.items():
            self.add_devices(device_string, number_of_threads)

        print('initial population:', self.devices)

        if '/' in name or '.' in name or '\\' in name or name in ['Documents', 'Pictures', 'Music', 'Public', 'Videos', 'Dokumente', 'Bilder', 'Musik', 'Downloads', 'Öffetnlich']:
            print('Error: For savety reasons some names and characters are forbidden to avoid the accidental removal of files or folders')
            devices = {}

        for device in self.devices:
            device.initialize()

        return

    def part_of_genome(self, small_genome, big_genome):
        result = True
        for key in small_genome:
            if key!='score' and key not in big_genome or small_genome[key] != big_genome[key]:
                result = False
        return result


    def new_score_event(self, genome):#called by devices
        found = None

        for g in self.running_individuals:
            if self.part_of_genome(g, genome):
                found = g

        #print('found', found)
        if found is not None:
            self.running_individuals.remove(found)
            found['score'] = genome['score']
            self.scored_individuals.append(found)
            self.Breed_And_Select.update_population()
            print('+', genome)#g
        else:
            self.error_event(genome, 'processed gene not found in running_individuals | running:' + str(self.running_individuals)+' scored:'+str(self.scored_individuals))

    def error_event(self, genome, message):#called by devices
        print('failed', message, genome)
        return #todo: implement

    def add_device(self, device_string):
        if device_string == 'single_thread':
            self.devices.append(Evolution_Device_Single_Thread(device_string, self))
        if device_string == 'multi_thread':
            self.devices.append(Evolution_Device_Multi_Thread(device_string, self))
        if 'ssh' in device_string:
            self.devices.append(Evolution_Device_SSH(device_string, self))

    def add_devices(self, device_string, number_of_threads):
        # move folder to remote device if needed
        for _ in range(number_of_threads):
            self.add_device(device_string)

        if len(self.devices)>0:
            self.devices[-1].initialize_device_group()

    def start(self):
        folder=get_data_folder()+'/StorageManager/'+self.name
        if not os.path.isdir(folder):

            for device in self.devices:
                device.start()

            self.active = True

            while self.active:
                for device in self.devices:
                    device.main_loop_update()
                time.sleep(1.0)

            for device in self.devices:
                device.stop()
            return True
        else:
            print('Warning:', folder, 'already exists. Remove folder or try continue() instead of start()')
            return False

    def stop(self):
        for device in self.devices:
            device.stop()

    def get_next_genome(self):
        result = None
        if len(self.non_scored_individuals) > 0:
            result = self.non_scored_individuals[0]
            self.non_scored_individuals.pop(0)
            self.running_individuals.append(result)

        #if result is None and len(self.running_individuals) > 0:
        #    result = self.running_individuals[0]

        if result is not None:
            result = result.copy()
            result['evo_name'] = self.name
            result['gen'] = self.Breed_And_Select.generation
            result['id'] = self.id_counter
            self.id_counter += 1
            result.update(self.inactive_genome_info)

        return result


from PymoNNto.Exploration.Network_UI.TabBase import *

class weight_tab(TabBase):

    def __init__(self, title='Weights', weight_attrs=['W']):
        super().__init__(title)
        self.weight_attrs = weight_attrs

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.weight_tab = Network_UI.Next_Tab(self.title)

        Network_UI.Add_element(QLabel('Min Max'), stretch=0)
        self.select_min_max_box = QComboBox()
        self.select_min_max_box.addItems(['Transmitter', 'Global', 'Variable', 'Single block'])
        self.select_min_max_box.setCurrentIndex(0)
        Network_UI.Add_element(self.select_min_max_box, stretch=4)
        Network_UI.Next_H_Block(stretch=0)

        #get max synapse group size
        max_sgs=2
        for group_tag in Network_UI.group_tags:
            for ng in Network_UI.network[group_tag]:
                for transmitter in Network_UI.transmitters:
                    l = 0
                    for weight_attr in self.weight_attrs:
                        l += len(Network_UI.get_combined_syn_mats(ng[transmitter], None, weight_attr))
                    max_sgs = np.maximum(max_sgs, l)

        self.transmitter_weight_images = {}
        for transmitter in Network_UI.transmitters:
            self.transmitter_weight_images[transmitter] = []
            for _ in range(max_sgs):
                image_item = Network_UI.Add_Image_Item(True, False, title='Neuron ' + transmitter + ' ?', tooltip_message='afferent synapse weights of selected neuron')#+self.weight_attr
                self.transmitter_weight_images[transmitter].append(image_item)
            Network_UI.Next_H_Block()

    def update(self, Network_UI):
        if self.weight_tab.isVisible() and len(Network_UI.network[Network_UI.neuron_select_group]) > 0:

            group = Network_UI.network[Network_UI.neuron_select_group, 0]
            #collect all synapse information
            syn_dict = {}

            # print(np.sum(group.W, axis=1))

            for transmitter in Network_UI.transmitters:
                syn_dict[transmitter] = {}
                for weight_attr in self.weight_attrs:

                    syns = Network_UI.get_combined_syn_mats(group.afferent_synapses[transmitter], Network_UI.neuron_select_id, weight_attr)

                    #print(np.sum(syns['GLU,syn,PC_1 => PC_1']), np.max(syns['GLU,syn,PC_1 => PC_1']))

                    if len(syns.values()) > 0:
                        syn_dict[transmitter][weight_attr]=syns


            #determine ranges
            min_max_mode = self.select_min_max_box.currentText()

            glob_max=0
            tr_max={}
            var_max={}
            key_max = {}

            for transmitter in syn_dict:
                tr_max[transmitter]=0
                var_max[transmitter] = {}
                key_max[transmitter] = {}
                for weight_attr in syn_dict[transmitter]:
                    var_max[transmitter][weight_attr] = 0
                    key_max[transmitter][weight_attr] = {}
                    for key in syn_dict[transmitter][weight_attr]:
                        m = np.max(syn_dict[transmitter][weight_attr][key])

                        tr_max[transmitter] = max(tr_max[transmitter], m)
                        var_max[transmitter][weight_attr] = max(var_max[transmitter][weight_attr], m)
                        key_max[transmitter][weight_attr][key] = m
                        glob_max = max(glob_max, m)

            if min_max_mode == 'Global':
                max_w = glob_max

            for transmitter in syn_dict:

                if min_max_mode == 'Transmitter':
                    max_w = tr_max[transmitter]

                for image, plot in self.transmitter_weight_images[transmitter]:
                    plot.setTitle('')
                    image.clear()

                i = 0
                for weight_attr in syn_dict[transmitter]:

                    if min_max_mode == 'Variable':
                        max_w = var_max[transmitter][weight_attr]

                    for key in syn_dict[transmitter][weight_attr]:

                        if min_max_mode == 'Single block':
                            max_w = key_max[transmitter][weight_attr][key]

                        self.transmitter_weight_images[transmitter][i][1].setTitle(key+' '+weight_attr) #1=plot
                        self.transmitter_weight_images[transmitter][i][0].setImage(np.rot90(syn_dict[transmitter][weight_attr][key], 3), levels=(0, max_w)) #0=image item
                        i += 1
                    #if weight_attr=='W_temp' and transmitter=='GLU':
                        #print(ss)




from PymoNNto.NetworkBehaviour.Input.Images.Helper import *
from PymoNNto.NetworkCore.Base import *
from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *

class Pixel_Pattern(PatternGroup):

    def initialize(self):
        self.patterns = []
        self.labels = []

        self.lastpattern = None
        self.pattern_show_duration_counter = 0
        self.pattern_duration_cycles = self.kwargs.get('pattern_duration_cycles', 1)

        self.individual_possibility = self.kwargs.get('individual_possibility', 0.0)
        self.overlapping_within_group = self.kwargs.get('overlapping_within_group', False)

        self.value_mode = self.kwargs.get('value_mode', 'fixed_value_mode')

        self.grid_width = self.kwargs.get('grid_width', 1)
        self.grid_height = self.kwargs.get('grid_height', 1)
        self.grid_channels = self.kwargs.get('grid_channels', 1)

        self.mask_mode = self.kwargs.get('mask_mode', 'center')

        self.select_mode = self.kwargs.get('select_mode', 'sequential')
        self.select_counter=-1

    def get_current_pattern_index(self):

        if self.select_mode == 'random':
            return np.random.randint(len(self.patterns))

        if self.select_mode == 'sequential':
            self.select_counter+=1
            if self.select_counter>=len(self.patterns):
                self.select_counter=0
            return self.select_counter

    def initialize_neuron_group(self, neurons):

        if self.mask_mode == 'first':
            neurons.Input_Mask = np.arange(neurons.size)<self.get_pattern_length()

        if self.mask_mode == 'random':
            chosen=np.random.choice(np.arange(neurons.size), self.get_pattern_length(), replace=False)
            neurons.Input_Mask = [(nr in chosen) for nr in np.arange(neurons.size)]

        if self.mask_mode == 'center':
            neurons.Input_Mask = (neurons.x < self.grid_width/2)*(neurons.x >= -self.grid_width/2)*(neurons.y < self.grid_height*self.grid_channels/2)*(neurons.y >= -self.grid_height*self.grid_channels/2)

    def get_pattern(self, neurons):

        self.current_pattern_index = -1

        if self.active:
            self.pattern_show_duration_counter -= 1
            if self.pattern_show_duration_counter > 0:
                return self.lastpattern

            result = None
            if len(self.patterns) > 0:
                result = neurons.get_neuron_vec()

                if self.overlapping_within_group:
                    self.current_pattern_index = []
                    for i, pattern in enumerate(self.patterns):
                        if np.random.rand() < self.individual_possibility:
                            result[neurons.Input_Mask] += self.get_pattern_values(pattern).flatten()
                            self.current_pattern_index.append(i)
                else:
                    r = self.get_current_pattern_index()
                    self.current_pattern_index = r
                    result[neurons.Input_Mask] += self.get_pattern_values(self.patterns[r]).flatten()

                self.lastpattern = result
                self.pattern_show_duration_counter = self.pattern_duration_cycles

            return result
        else:
            return None

    def get_pattern_values(self, pattern):
        if pattern.shape != self.get_pattern_dimension():
            x = np.random.randint(pattern.shape[2] - self.grid_width)
            y = np.random.randint(pattern.shape[1] - self.grid_height)
            return pattern[:, y:y + self.grid_height, x:x + self.grid_width]
        else:
            return pattern

    def get_vstackpattern_dimension(self):
        dim = self.get_pattern_dimension()
        return (dim[0] * dim[1], dim[2])

    def get_pattern_dimension(self):
        return (self.grid_channels, self.grid_height, self.grid_width)

    def get_pattern_length(self):
        return self.grid_channels*self.grid_height*self.grid_width

    def reconstruct_pattern(self, p):
        return p.reshape(self.get_pattern_dimension())


from PymoNNto.Exploration.Network_UI.TabBase import *


class sidebar_fast_forward_module(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def recorder_on_off(self, Network_UI, enable):
        for rec in Network_UI.network['UI_rec']:
            rec.behaviour_enabled = enable or self.rec_cb.isChecked()

    def initialize(self, Network_UI):

        def start_pause_click(event):
            Network_UI.pause = not Network_UI.pause

        self.sp_btn = QPushButton('start/pause', Network_UI.main_window)
        self.sp_btn.clicked.connect(start_pause_click)
        Network_UI.Add_Sidebar_Element(self.sp_btn)

        self.rec_cb = QCheckBox()
        self.rec_cb.setText('fast forward record')
        self.rec_cb.setChecked(False)
        Network_UI.Add_Sidebar_Element(self.rec_cb)

        h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

        Network_UI.step = False

        def one_step(event):
            Network_UI.step = True

        self.os_btn = QPushButton('1 step', Network_UI.main_window)
        self.os_btn.clicked.connect(one_step)
        # self.Add_Sidebar_Element(self.os_btn)
        h_layout.addWidget(self.os_btn)

        def fast_forward(event):
            self.recorder_on_off(Network_UI, False)
            Network_UI.network.simulate_iterations(100, 100, measure_block_time=True)
            self.recorder_on_off(Network_UI, True)

        self.ff_btn = QPushButton('100', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        def fast_forward(event):
            self.recorder_on_off(Network_UI, False)
            Network_UI.network.simulate_iterations(1000, 100, measure_block_time=True)
            self.recorder_on_off(Network_UI, True)

        self.ff_btn = QPushButton('1k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

        def fast_forward(event):
            self.recorder_on_off(Network_UI, False)
            Network_UI.network.simulate_iterations(5000, 100, measure_block_time=True)
            self.recorder_on_off(Network_UI, True)

        self.ff_btn = QPushButton('5k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        def fast_forward(event):
            self.recorder_on_off(Network_UI, False)
            Network_UI.network.simulate_iterations(15000, 100, measure_block_time=True)
            self.recorder_on_off(Network_UI, True)

        self.ff_btn = QPushButton('15k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        def fast_forward(event):
            self.recorder_on_off(Network_UI, False)
            Network_UI.network.simulate_iterations(50000, 100, measure_block_time=True)
            self.recorder_on_off(Network_UI, True)

        self.ff_btn = QPushButton('50k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

    def update(self, Network_UI):
        return


from PymoNNto.NetworkCore.Neuron_Group import *


class NeuronManualActivator(Behaviour):

    #def __init__(self, clip_min=0.0, clip_max=1.0, write_to='glu_inter_gamma_activity'):
    #    self.clip_min = clip_min
    #    self.clip_max = clip_max
    #    self.write_to = write_to

    def set_variables(self, neurons):
        self.clip_min = self.get_init_attr('clip_min', 0.0, neurons)
        self.clip_max = self.get_init_attr('clip_max', 1.0, neurons)
        self.write_to = self.get_init_attr('write_to', 'glu_inter_gamma_activity', neurons)

        self.neurons=neurons
        neurons.outside_activation = neurons.get_neuron_vec()

        if not hasattr(neurons, self.write_to):
            setattr(neurons, self.write_to, neurons.get_neuron_vec())

    def new_iteration(self, neurons):
        getattr(neurons, self.write_to)[:] *= 0
        getattr(neurons, self.write_to)[:] += neurons.outside_activation
        getattr(neurons, self.write_to)[:] = np.clip(getattr(neurons, self.write_to)[:], self.clip_min, self.clip_max)
        neurons.outside_activation *= 0.0

    def activate(self, act, additive=False):
        if additive:
            self.neurons.outside_activation += np.array(act)
        else:
            self.neurons.outside_activation = np.array(act)


class NeuronActivator(Behaviour):

    #def __init__(self, write_to='glu_inter_gamma_activity', pattern_groups=None, clip_min=0.0, clip_max=1.0):

    def find_objects(self, key):
        result = []
        for pg in self.TNAPatterns:
            result += pg[key]
        return result


    def set_variables(self, neurons):
        self.add_tag('activator')

        neurons.pattern_index = 0

        pattern_groups = self.get_init_attr('pattern_groups', None, neurons)

        self.clip_min = self.get_init_attr('clip_min', 0.0, neurons)
        self.clip_max = self.get_init_attr('clip_max', 1.0, neurons)
        self.write_to = self.get_init_attr('write_to', 'glu_inter_gamma_activity', neurons)

        setattr(neurons, self.write_to, neurons.get_neuron_vec())

        self.TNAPatterns = []

        self.active=True

        if pattern_groups is not None:
            self.add_patternGroups(pattern_groups, neurons)

        #self.neurons=neurons
        #if not hasattr(neurons, self.write_to):
        #    setattr(neurons, self.write_to, neurons.get_neuron_vec())

    def get_pattern_samples(self, number, size):
        result = []
        for i in range(number):
            result.append(self.get_pattern_values())
        return np.array(result)



    def get_pattern_values(self, neurons=None):
        patterns = []
        overlapping = np.zeros(neurons.size)

        if self.active:
            for pattern_group in self.TNAPatterns:

                if pattern_group.group_possibility >= 1.0 or np.random.rand() < pattern_group.group_possibility:
                    pat = pattern_group.get_pattern(neurons)
                    if pat is not None:
                        return pat
                else:
                    pat = None

                if pat is not None:
                    #if pat.shape is not (neurons.size,):
                    #    pat.resize((neurons.size), refcheck=False)
                    if pattern_group.overlapping_with_other_groups:
                        overlapping += pat
                    else:
                        if pattern_group.has_overlapping_priority:
                            return pat
                        else:
                            patterns.append(pat)

        patterns.append(overlapping)
        r = np.random.randint(len(patterns))
        return patterns[r]


    def new_iteration(self, neurons):
        #if self.active:
        #getattr(neurons, self.write_to)[:] *= 0 #todo test: moved from outside "if" block inside it
        getattr(neurons, self.write_to)[:] += self.get_pattern_values(neurons)
        neurons.pattern_index = self.TNAPatterns[0].current_pattern_index#todo: better
        #neurons.input_activity += self.get_pattern_values()
        #neurons.input_activity = np.clip(neurons.input_activity, self.clip_min, self.clip_max)#neccessary? multiple clipped

    def add_patternGroups(self, patterns, neurons):
        #if not hasattr(self, 'TNAPatterns'):
        #    self.TNAPatterns = []

        if type(patterns) == list:
            for p in patterns:
                self.TNAPatterns.append(p)
                p.initialize_neuron_group(neurons)
        else:
            self.TNAPatterns.append(patterns)
            patterns.initialize_neuron_group(neurons)





class NeuronPreprocessedActivator(NeuronActivator):

    def preprocess(self,preprocessing_steps, neurons):
        self.pattern_list = []
        for _ in range(preprocessing_steps):
            self.pattern_list.append(super().get_pattern_values(neurons))
            self.step = 0

    def reset(self):
        self.step = 0

    def get_pattern_values(self, neurons):
        self.step += 1
        if self.step<len(self.pattern_list):
            return self.pattern_list[self.step-1]
        else:
            return super().get_pattern_values(neurons)



from PymoNNto.Exploration.Network_UI.TabBase import *
import random

class single_group_plot_tab(TabBase):

    def __init__(self, variables={'output', 'excitation', 'inhibition', 'input_act', 'TH'}, title='default group', timesteps=500):
        super().__init__(title)
        self.variables = {}
        for i, v in enumerate(variables):
            self.variables[v] = self.get_color(i)
        self.timesteps=timesteps

    def get_color(self, i):
        colors = [(0, 0, 0), (0, 0, 255), (255, 0, 0), (255, 0, 255), (0, 255, 0)]
        if i <= len(colors):
            return colors[i]
        else:
            return (random.randint(0,255),random.randint(0,255),random.randint(0,255))


    def add_recorder_variables(self, neuron_group, Network_UI):

        for var in self.variables:
            if hasattr(neuron_group, var):
                Network_UI.add_recording_variable(neuron_group, 'n.'+var, timesteps=self.timesteps)
                Network_UI.add_recording_variable(neuron_group, 'np.mean(n.'+var+')', timesteps=self.timesteps)

    def initialize(self, Network_UI):
        self.inh_exc_tab = Network_UI.Next_Tab(self.title)


        tooltip_message = ''#str(self.variables).replace(',', '\r\n')
        for key in self.variables:
            tooltip_message+=str(self.variables[key])+' : '+str(key)+'\r\n'

        single_vars = list(self.variables.keys())
        mean_vars = ['np.mean(n.'+var+')' for var in self.variables]
        colors = list(self.variables.values())

        mean_curves = Network_UI.Add_plot_curve('NeuronGroup' + str(single_vars), number_of_curves=len(mean_vars), names=mean_vars, colors=colors, lines=[0], tooltip_message=tooltip_message, return_list=True)
        self.net_curves=dict(zip(self.variables, mean_curves))

        Network_UI.Next_H_Block()

        single_curves = Network_UI.Add_plot_curve('Neuron' + str(single_vars), number_of_curves=len(single_vars), names=single_vars, colors=colors, lines=[0], tooltip_message=tooltip_message, return_list=True)
        self.neuron_curves = dict(zip(self.variables, single_curves))

        #Network_UI.Next_H_Block()


    def update(self, Network_UI):
        if self.inh_exc_tab.isVisible():
            group = Network_UI.network[Network_UI.neuron_select_group, 0]

            #rec = Network_UI.rec(group, self.timesteps)

            for var in self.variables:
                try:
                #if hasattr(group, var):
                    iterations = group['n.iteration', 0, 'np'][-self.timesteps:]

                    mean_var = 'np.mean(n.'+var+')'
                    single_var = 'n.'+var

                    data_mean = group[mean_var, 0, 'np'][-self.timesteps:]
                    self.net_curves[var].setData(iterations, data_mean)

                    data_neuron = group[single_var, 0, 'np'][-self.timesteps:, Network_UI.neuron_select_id]
                    self.neuron_curves[var].setData(iterations, data_neuron)
                except:
                    print(var, "cannot be evaluated")



from PymoNNto.NetworkCore.Network import *
from PymoNNto.NetworkCore.Neuron_Group import *
from PymoNNto.NetworkCore.Synapse_Group import *
from PymoNNto.NetworkBehaviour.Structure.Structure import *
from PymoNNto.NetworkBehaviour.Recorder.Recorder import *

#from PymoNNto.NetworkBehaviour.Structure.Structure import get_squared_dim

from PymoNNto.Exploration.StorageManager.StorageManager import *





from PymoNNto.NetworkBehaviour.Logic.Basics.BasicHomeostasis import *
from PymoNNto.NetworkBehaviour.Logic.Basics.Normalization import *

class HomeostaticMechanism(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Homeostasis')
        self.ACTIVITY_COUNT_TH_MIN = self.get_init_attr('ACTIVITY_COUNT_TH_MIN', 0.0, neurons)#step height #neurons.firetreshold
        self.ACTIVITY_COUNT_TH_MAX = self.get_init_attr('ACTIVITY_COUNT_TH_MAX', 1.0, neurons)

        self.pattern_chance = self.get_init_attr('pattern_chance', 0.01, neurons)
        self.min = self.get_init_attr('min', self.pattern_chance, neurons)#*self.ACTIVITY_COUNT_TH_MIN
        self.max = self.get_init_attr('max', self.pattern_chance, neurons)#*self.ACTIVITY_COUNT_TH_MAX
        self.range_end = self.get_init_attr('range_end', 1000, neurons)
        self.inc = self.get_init_attr('inc', 0.01, neurons)
        self.dec = self.get_init_attr('dec', 0.01, neurons)

        self.target_min = self.get_init_attr('target_min', 0.1, neurons)#0.01
        self.target_max = self.get_init_attr('target_max', 100, neurons)#100000

        self.target_attr = self.get_init_attr('target_attr', 'weight_norm_factor', neurons)
        self.src_attr = self.get_init_attr('src_attr', 'output_activity_history[0]', neurons)

        self.avg = neurons.get_random_neuron_vec()*(self.max-self.min)+self.min

    def new_iteration(self, neurons):
        if neurons.learning:
            current_value = neurons.output_activity_history[0].copy()
            #current_value = getattr(neurons, self.src_attr).copy()#neurons.output_activity_history[0].copy() #input_activity_history  #COPY!!!

            if self.ACTIVITY_COUNT_TH_MIN != 0.0: current_value *= current_value > self.ACTIVITY_COUNT_TH_MIN
            if self.ACTIVITY_COUNT_TH_MAX != 1.0: current_value *= current_value < self.ACTIVITY_COUNT_TH_MAX

            self.avg = (self.avg * self.range_end + current_value) / (self.range_end+1)

            greater = (self.avg > self.max) * (-self.dec) * (self.avg-self.max)#/self.val_avg.range_end
            smaller = (self.avg < self.min) * (self.inc) * (self.min-self.avg)#/self.val_avg.range_end

            if self.target_attr != '':
                setattr(neurons, self.target_attr, np.clip(getattr(neurons, self.target_attr)+greater+smaller, self.target_min, self.target_max))
            #else:
            #    print(neurons.avg)

#class GABANormalization(TRENNeuronBehaviour):

#    def set_variables(self, neurons):
#        #neurons.GABA_norm_value = neurons.get_random_neuron_vec()*self.get_init_attr('GABA_norm_value', 1.0)
#        neurons.GABA_norm_value = neurons.get_neuron_vec() + self.get_init_attr('GABA_norm_value', 1.0, neurons)

#        #neurons.GABA_norm_value[0] = 10.0

#    def new_iteration(self, neurons):
#        if neurons.learning:
#            neurons.temp_weight_sum = neurons.get_neuron_vec()

#            for s in neurons.afferent_synapses['GABA']:
#                s.dst.temp_weight_sum += np.sum(s.GABA_Synapses, axis=1)
                #s.get_dest_vec('temp_weight_sum')[:] += np.sum(s.GABA_Synapses, axis=1)

#            neurons.temp_weight_sum /= neurons.GABA_norm_value

#            for s in neurons.afferent_synapses['GABA']:
#                #s.GABA_Synapses /= s.get_dest_vec('temp_weight_sum')[:, None]
#                #s.GABA_Synapses = np.clip(s.GABA_Synapses, 0, None)
#                print(s.get_dest_vec('temp_weight_sum')[:, None])



class GlutamateCacheConvergeAndNormalization(Behaviour):

    def set_variables(self, neurons):
        self.add_tag('Normalization')
        nv = self.get_init_attr('norm_value', 2.7, neurons)
        #neurons.norm_value = neurons.get_neuron_vec() + self.get_init_attr('norm_value', 2.7)
        #neurons.norm_value = neurons.get_random_neuron_vec()*self.get_init_attr('norm_value', 2.7)
        #neurons.norm_value = neurons.get_random_neuron_vec()*nv+0.0001
        neurons.weight_norm_factor = (neurons.get_neuron_vec()+1.0) * nv

        for s in neurons.afferent_synapses['GLU']:
            s.W = s.get_synapse_mat()

        self.new_iteration(neurons)


    def new_iteration(self, neurons):
        if neurons.learning:
            #neurons.temp_weight_sum = neurons.get_neuron_vec()

            for s in neurons.afferent_synapses['GLU']:
                s.W = np.sum(s.W_Caches, axis=0)

            normalize_synapse_attr('W', 'W_Caches', neurons.weight_norm_factor, neurons, 'GLU')

        for s in neurons.afferent_synapses['GLU']:
            s.W = np.clip(np.sum(s.W_Caches, axis=0), 0, None)#todo move to separate behaviour and do construction there


#only for LGN

class TREN_input_weighting(Behaviour):

    def set_variables(self, neurons):
        neurons.norm_value = neurons.get_random_neuron_vec()*0.5#0.4-0.2+0.3 #neurons.get_neuron_vec()+1

    def new_iteration(self, neurons):
        neurons.glu_inter_gamma_activity *= neurons.norm_value


from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Analysis.PCA import *

class PCA_tab(TabBase):

    def __init__(self, parameter='output', title='PCA', timesteps=1000):
        super().__init__(title)
        self.parameter = parameter
        self.timesteps = timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        if hasattr(neuron_group, self.parameter):
            Network_UI.add_recording_variable(neuron_group, 'n.'+self.parameter, timesteps=self.timesteps)

    def initialize(self, Network_UI):
        self.pca_tab = Network_UI.Next_Tab(self.title)

        self.pca_curve = Network_UI.Add_plot_curve('PCA singular values', colors=[(0, 0, 0)], legend=False, x_label='singular value', y_label='')

        self.evr_curve = Network_UI.Add_plot_curve('PCA explained variance ratio', colors=[(0, 0, 0)], legend=False, x_label='features', y_label='% variance explained')

        Network_UI.Next_H_Block()

        self.singular_value_count=5

        self.singular_value_histories = [[] for i in range(self.singular_value_count)]
        self.singular_value_iterations = []

        self.sv_curves = Network_UI.Add_plot_curve(str(self.singular_value_count)+' biggest singular values', number_of_curves=self.singular_value_count, legend=False,x_label='singular value', y_label='')


    def update(self, Network_UI):
        if self.pca_tab.isVisible() and Network_UI.network.iteration%10==0:
            if len(Network_UI.network[Network_UI.neuron_select_group]) >= 0:
                group=Network_UI.network[Network_UI.neuron_select_group, 0]

                try:
                #if hasattr(group, self.parameter):

                    act = group['n.'+self.parameter, 0, 'np'][-self.timesteps:]
                    try :
                        pca = get_PCA(act, 100)
                        svs = pca.singular_values_
                        evr = pca.explained_variance_ratio_
                        if len(svs) >= self.singular_value_count:
                            self.singular_value_iterations.append(Network_UI.network.iteration)
                            for i in range(self.singular_value_count):
                                self.singular_value_histories[i].append(svs[i])
                                self.sv_curves[i].setData(self.singular_value_iterations, self.singular_value_histories[i])

                            self.pca_curve.setData(svs)
                            self.evr_curve.setData(np.cumsum(np.round(evr, decimals=3)*100))
                    except:
                        print('not enough timesteps for PCA. waiting...')

                except:
                    print(self.parameter, "cannot be evaluated")




'''
from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.NetworkBehaviour.Recorder.Recorder import *

class individual_weight_tab(TabBase):

    def __init__(self, title='Ind Weights', weight_attr='W'):
        super().__init__(title)
        self.weight_attr = weight_attr

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.individual_weight_tab = Network_UI.Next_Tab(self.title)

        self.plot = Network_UI.Add_plot('Neuron Weights', stretch = 8)

        self.rec_name=self.title + '_rec'
        self.Network_UI=Network_UI

        self.current_ng=None
        self.neuron_index=-1
        self.my_syn_rec=None

        self.listWidget = QListWidget()
        Network_UI.Add_element(self.listWidget)

        Network_UI.Next_H_Block()

        self.min_w_slider = QSlider(1)  # QtCore.Horizontal
        self.min_w_slider.setMinimum(-1)
        self.min_w_slider.setMaximum(100)
        self.min_w_slider.setSliderPosition(-1)
        self.min_w_slider.setToolTip('slide to cut away x% of smallest weights')
        self.min_w_slider.mouseReleaseEvent = Network_UI.static_update_func
        Network_UI.Add_element(self.min_w_slider)


    def start_rec(self, group):

        # self.key = 'np.concatenate([s.W[:,np.where(s.dst.id=='+str(self.neuron_index)+')[0]][s.enabled[:,np.where(s.dst.id=='+str(self.neuron_index)+')[0]]] for s in n.afferent_synapses["All"]])'
        # self.my_syn_rec = Recorder([self.key], tag=self.rec_name)

        if len(group[self.rec_name]) == 0 or self.Network_UI.neuron_select_id != self.neuron_index:

            if self.current_ng is not None and len(self.current_ng[self.rec_name]) > 0:
                self.Network_UI.network.remove_behaviours_from_object(self.current_ng, tags=[self.rec_name])
                self.my_syn_rec=None
                print('stop')

            #if self.my_syn_rec is not None:
            #    self.Network_UI.network.remove_behaviours_from_neuron_groups([self.current_ng], tags=[self.rec_name])
            #        print('stop')

            print('started', group)
            self.plot.clear()

            self.neuron_index = self.Network_UI.neuron_select_id

            self.curves = {}
            self.keys = []
            self.checkboxes={}
            self.listWidget.clear()
            for i, s in enumerate(group.afferent_synapses["All"]):
                new_id = np.where(s.dst.id == self.neuron_index)[0]
                if len(new_id) > 0:# and hasattr(s, self.weight_attr): #TODO:
                    new_id = new_id[0]

                    weight = eval('s.' + self.weight_attr)

                    if type(s.enabled)==bool:
                        syn_count = len(weight[new_id])
                    else:
                        syn_count = len(weight[new_id,:][s.enabled[new_id,:]])

                    #print(weight.shape, new_id, enabled)
                    #syn_count = len(weight[new_id][enabled])#,:
                    if syn_count > 0:

                        color = (s.src.color[0], s.src.color[1], s.src.color[2], 255)

                        if type(group.afferent_synapses["All"][i].enabled) == bool:
                            enabled_string = 'True'
                        else:
                            enabled_string='n.afferent_synapses["All"]['+str(i)+'].enabled['+str(new_id)+',:]'


                        k = 'n.afferent_synapses["All"]['+str(i)+'].W['+str(new_id)+',:]['+enabled_string+']'
                        self.curves[k] = []
                        self.keys.append(k)

                        self.checkboxes[k] = QListWidgetItem(str(s.src.tags[0]))
                        self.checkboxes[k].setCheckState(2)
                        self.listWidget.addItem(self.checkboxes[k])

                        for i in range(syn_count):
                            curve = pg.PlotCurveItem([], pen=color)  # pen=colors[i%len(colors)]
                            self.plot.addItem(curve)
                            self.curves[k].append(curve)

            self.my_syn_rec = Recorder(self.keys + ['n.iteration'], tag=self.rec_name, gapwidth=10)
            self.Network_UI.network.add_behaviours_to_object({10001: self.my_syn_rec}, group)
            self.current_ng = group


            #size = len(np.concatenate([s.W[:, new_indx][s.enabled[:, new_indx]] for s in group.afferent_synapses["All"]]))
            #print(size)

            return False
        else:
            return True

    def update(self, Network_UI):

        if self.individual_weight_tab.isVisible():
            if len(Network_UI.network[Network_UI.neuron_select_group]) > 0:

                group = Network_UI.network[Network_UI.neuron_select_group, 0]

                if (self.start_rec(group) and self.my_syn_rec.is_new_data_available()) or self.my_syn_rec.active == False:

                    if self.my_syn_rec.active==False:
                        print('continue')
                        self.my_syn_rec.active=True

                    x_data=self.my_syn_rec['n.iteration', 0, 'np']

                    for k in self.keys:

                        neuron_data = self.my_syn_rec[k, 0, 'np']

                        #print(neuron_data)

                        if len(neuron_data)>0:
                            mws = np.max(neuron_data)/100*self.min_w_slider.sliderPosition()

                            for i, curve in enumerate(self.curves[k]):
                                if self.checkboxes[k].checkState() == 2:
                                    d=neuron_data[:, i]
                                    #if d[-1]>=mws:
                                    #print(x_data.shape,d.shape)
                                    curve.setData(x_data, d)
                                    #else:
                                    #    curve.clear()
                                else:
                                    curve.clear()
        else:
            if self.my_syn_rec is not None and self.my_syn_rec.active==True:
                print('paused')
                self.my_syn_rec.active=False
'''


from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.NetworkBehaviour.Recorder.Recorder import *

class individual_weight_tab(TabBase):

    def __init__(self, title='Ind Weights', weight_attr='W'):
        super().__init__(title)
        self.weight_attr = weight_attr

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.individual_weight_tab = Network_UI.Next_Tab(self.title)

        self.plot = Network_UI.Add_plot('Neuron Weights', stretch = 8)

        self.rec_name=self.title + '_rec'
        self.Network_UI=Network_UI

        self.current_ng=None
        self.neuron_index=-1
        self.my_syn_rec=None

        self.key=''

        self.listWidget = Network_UI.Add_element(QListWidget())

        Network_UI.Next_H_Block()

        self.min_w_slider = QSlider(1)  # QtCore.Horizontal
        self.min_w_slider.setMinimum(-1)
        self.min_w_slider.setMaximum(100)
        self.min_w_slider.setSliderPosition(-1)
        self.min_w_slider.setToolTip('slide to cut away x% of smallest weights')
        self.min_w_slider.mouseReleaseEvent = Network_UI.static_update_func
        Network_UI.Add_element(self.min_w_slider)


    def start_rec(self, group):

        #if self.current_ng is not None and len(self.current_ng[self.rec_name]) > 0 and self.Network_UI.neuron_select_id != self.neuron_index:
        #    self.Network_UI.network.remove_behaviours_from_object(self.current_ng, tags=[self.rec_name])
        #    self.my_syn_rec = None
        #    print('stop')

        if self.current_ng is not None and (self.neuron_index != self.Network_UI.neuron_select_id or self.current_ng != group):

            for s in self.current_ng.afferent_synapses["All"]:
                s.UI_recorder = None
                s.cb = None
                self.Network_UI.network.remove_behaviours_from_object(s, tags=[self.key])

            self.listWidget.clear()
            self.plot.clear()

            self.current_ng = None
            self.neuron_index = -1
            print('removed')

        if self.current_ng is None:#len(self.Network_UI.network[self.key]) == 0:

            self.current_ng = group
            self.neuron_index = self.Network_UI.neuron_select_id

            # self.key = 'np.concatenate([s.W[:,np.where(s.dst.id=='+str(self.neuron_index)+')[0]][s.enabled[:,np.where(s.dst.id=='+str(self.neuron_index)+')[0]]] for s in n.afferent_synapses["All"]])'
            # self.my_syn_rec = Recorder([self.key], tag=self.rec_name)


            #if self.my_syn_rec is not None:
            #    self.Network_UI.network.remove_behaviours_from_neuron_groups([self.current_ng], tags=[self.rec_name])
            #        print('stop')

            print('started')
            self.plot.clear()

            self.key = 's.'+self.weight_attr+'['+str(self.neuron_index)+']'

            self.checkboxes={}
            self.listWidget.clear()

            for s in group.afferent_synapses["All"]:
                s.UI_recorder = Recorder([self.key, 's.iteration'], tag=self.rec_name, gapwidth=10, save_as_numpy=True)
                self.Network_UI.network.add_behaviours_to_object({10001: s.UI_recorder}, s)

                s.UI_curves = []
                syn_data = eval(self.key)

                s.cb = QListWidgetItem(str(s.tags[0]))
                s.cb.setCheckState(2)
                self.listWidget.addItem(s.cb)

                for i in range(len(syn_data)):
                    if (type(s.enabled) is bool and s.enabled) or s.enabled[self.neuron_index, i]:
                        color = (s.src.color[0], s.src.color[1], s.src.color[2], 255)
                        curve = pg.PlotCurveItem([], pen=color)  # pen=colors[i%len(colors)]
                        curve.index = i
                        self.plot.addItem(curve)
                        s.UI_curves.append(curve)


                #self.curves[k] = []
                #self.keys.append(k)

                #self.checkboxes[k] = QListWidgetItem(str(s.src.tags[0]))
                #self.checkboxes[k].setCheckState(2)
                #self.listWidget.addItem(self.checkboxes[k])

                #for i in range(syn_count):


                '''
                new_id = np.where(s.dst.id == self.neuron_index)[0]
                if len(new_id) > 0:# and hasattr(s, self.weight_attr): #TODO:
                    new_id = new_id[0]

                    weight = eval('s.' + self.weight_attr)

                    if type(s.enabled)==bool:
                        syn_count = len(weight[new_id])
                    else:
                        syn_count = len(weight[new_id,:][s.enabled[new_id,:]])

                    #print(weight.shape, new_id, enabled)
                    #syn_count = len(weight[new_id][enabled])#,:
                    if syn_count > 0:

                        color = (s.src.color[0], s.src.color[1], s.src.color[2], 255)

                        if type(group.afferent_synapses["All"][i].enabled) == bool:
                            enabled_string = 'True'
                        else:
                            enabled_string='n.afferent_synapses["All"]['+str(i)+'].enabled['+str(new_id)+',:]'


                        k = 'n.afferent_synapses["All"]['+str(i)+'].W['+str(new_id)+',:]['+enabled_string+']'
                        self.curves[k] = []
                        self.keys.append(k)

                        self.checkboxes[k] = QListWidgetItem(str(s.src.tags[0]))
                        self.checkboxes[k].setCheckState(2)
                        self.listWidget.addItem(self.checkboxes[k])

                        for i in range(syn_count):
                            curve = pg.PlotCurveItem([], pen=color)  # pen=colors[i%len(colors)]
                            self.plot.addItem(curve)
                            self.curves[k].append(curve)
                '''

            #self.my_syn_rec = Recorder(self.keys + ['n.iteration'], tag=self.rec_name, gapwidth=10)
            #self.Network_UI.network.add_behaviours_to_object({10001: self.my_syn_rec}, group)
            #self.current_ng = group


            #size = len(np.concatenate([s.W[:, new_indx][s.enabled[:, new_indx]] for s in group.afferent_synapses["All"]]))
            #print(size)


    def update(self, Network_UI):

        if self.individual_weight_tab.isVisible():
            if len(Network_UI.network[Network_UI.neuron_select_group]) > 0:

                group = Network_UI.network[Network_UI.neuron_select_group, 0]
                self.start_rec(group)

                for s in self.current_ng.afferent_synapses["All"]:
                    if not s.UI_recorder.active:
                        s.UI_recorder.active = True
                        print('recorder started')
                    if s.cb.checkState() == 2:
                        if s.UI_recorder.is_new_data_available():
                            x_data = s.UI_recorder['s.iteration', 0]
                            y_data = s.UI_recorder[self.key, 0]

                            if len(x_data) > 0 and len(y_data) > 0:
                                for curve in s.UI_curves:
                                    curve.setData(x_data, y_data[:, curve.index])
        else:
            if self.current_ng is not None:
                for s in self.current_ng.afferent_synapses["All"]:
                    if s.UI_recorder.active:
                        s.UI_recorder.active=False
                        print('recorder paused')



from PymoNNto.Exploration.Evolution.Interface_Functions import *

print(sys.argv)

print('slave')

print('genes: ', get_genome())
print('gene a: ', get_gene('a', None))
print('gene b: ', gene('b', 1))

score = get_gene('a', 0) + get_gene('b', 0)

set_score(score)



from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *

class Noise_Pattern(PatternGroup):

    def initialize(self):
        self.density = self.kwargs.get('density', 0.1)
        self.binary = self.kwargs.get('binary', True)
        self.max = self.kwargs.get('max', 1)
        self.min = self.kwargs.get('min', 0)

    def get_pattern(self, neurons):
        result=(np.random.rand(self.grid_channels, self.grid_height, self.grid_width)<self.density).astype(def_dtype)
        if not self.binary:
            result*=np.random.rand(self.size)
        return self.min+result*self.max



from PymoNNto.Exploration.Evolution.Evolution import *

if __name__ == '__main__':
    genome = {'a': 1, 'b': 2}

    evo = Evolution(name='my_test_evo',
                    slave_file='test_slave.py', #Exploration/Evolution/
                    individual_count=5,
                    mutation=0.4,
                    death_rate=0.5,
                    constraints=['b > a+1', 'a<=1', 'b>= 2'],
                    inactive_genome_info={'info': 'my_info'},
                    ui=False,
                    start_genomes=[genome],
                    devices={'single_thread': 0,
                             'multi_thread': 0,
                             'ssh vieth@poppy.fias.uni-frankfurt.de': 0,
                             'ssh marius@hey3kmuagjunsk2b.myfritz.net': 1,
                             }
                    )

    evo.start()



from PymoNNto.NetworkCore.Neuron_Group import *
import math
import numpy as np


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

def get_squared_dim(number_of_neurons, depth=1, round_up=True):

    divider=int(np.trunc(np.sqrt(number_of_neurons)))

    while divider>1 and number_of_neurons%divider!=0:
      divider-=1

    w=divider
    h=int(number_of_neurons/divider)

    print(w,h)
    return NeuronDimension(width=w, height=h, depth=depth)


class NeuronDimension(Behaviour):

    run_on_init = True

    #def __init__(self, width=1, height=1, depth=1):
    #    super().__init__()

    def set_positions(self, width=1, height=1, depth=1):
        self.neurons.x = np.array([i%width for i in range(self.neurons.size)]).astype(def_dtype)
        self.neurons.y = np.array([np.floor(i/width)%height for i in range(self.neurons.size)]).astype(def_dtype)
        self.neurons.z = np.array([np.floor(i/(width*height)) for i in range(self.neurons.size)]).astype(def_dtype)

    def apply_pattern_transformation_function(self, transform_mat, hup, wup, depth):

        big_x_mat = np.tile(np.arange(wup), (1, hup)).reshape(1, hup, wup)
        big_y_mat = np.tile(np.arange(wup), (1, hup)).reshape(1, hup, wup).transpose()

        big_x_mat = np.repeat(big_x_mat[:, :, :], depth, axis=0).flatten()
        big_y_mat = np.repeat(big_y_mat[:, :, :], depth, axis=0).flatten()

        self.neurons.x = big_x_mat[transform_mat]
        self.neurons.y = big_y_mat[transform_mat]

    def move(self, x=0, y=0, z=0):
        self.neurons.x += x
        self.neurons.y += y
        self.neurons.z += z
        return self

    def scale(self, x=1, y=1, z=1):
        self.neurons.x *= x
        self.neurons.y *= y
        self.neurons.z *= z
        return self

    def noise(self, x_noise=1, y_noise=1, z_noise=1, centered=True):
        self.neurons.x += np.random.rand(len(self.neurons.x)) * x_noise
        self.neurons.y += np.random.rand(len(self.neurons.y)) * y_noise
        self.neurons.z += np.random.rand(len(self.neurons.z)) * z_noise

        if centered:
            self.neurons.x -= x_noise / 2
            self.neurons.y -= y_noise / 2
            self.neurons.z -= z_noise / 2
        return self

    def rotate(self, axis, theta):
        for i in range(self.size):
            self.neurons.x[i], self.neurons.y[i], self.neurons.z[i] = np.dot(rotation_matrix(axis, theta), [self.neurons.x[i], self.neurons.y[i], self.neurons.z[i]])
        return self

    #def stretch_to_equal_size(self, target_neurons):
    #    if hasattr(target_neurons, 'width') and self.neurons.width>0:
    #        x_stretch = target_neurons.width/self.neurons.width
    #        self.neurons.x *= x_stretch

    #    if hasattr(target_neurons, 'height') and self.neurons.height > 0:
    #        y_stretch = target_neurons.height/self.neurons.height
    #        self.neurons.y *= y_stretch

    #    if hasattr(target_neurons, 'depth') and self.neurons.depth > 0:
    #        z_stretch = target_neurons.depth/self.neurons.depth
    #        self.neurons.z *= z_stretch

    def stretch_to_equal_size(self, target_neurons):

        if hasattr(target_neurons, 'width') and self.neurons.width>0:
            x_stretch = target_neurons.width/self.neurons.width
            self.neurons.x *= x_stretch

        if hasattr(target_neurons, 'height') and self.neurons.height > 0:
            y_stretch = target_neurons.height/self.neurons.height
            self.neurons.y *= y_stretch

        if hasattr(target_neurons, 'depth') and self.neurons.depth > 0:
            z_stretch = target_neurons.depth/self.neurons.depth
            self.neurons.z *= z_stretch


    def set_variables(self, neurons):
        self.add_tag('structure')

        self.width = self.get_init_attr('width', 1, neurons)
        self.height = self.get_init_attr('height', 1, neurons)
        self.depth = self.get_init_attr('depth', 1, neurons)

        for pg in self.get_init_attr('input_pattern_groups', [], neurons):
            dim = pg.get_pattern_dimension()#.reconstruct_pattern(None)
            if len(dim) > 0: self.height = np.maximum(self.height, dim[0])
            if len(dim) > 1: self.width = np.maximum(self.width, dim[1])
            if len(dim) > 2:
                self.depth = np.maximum(self.depth, dim[0])
                self.height = np.maximum(self.height, dim[1])
                self.width = np.maximum(self.width, dim[2])

        self.neurons=neurons
        neurons.width = self.width
        neurons.height = self.height
        neurons.depth = self.depth
        neurons.size=self.width*self.height*self.depth
        self.set_positions(self.width, self.height, self.depth)
        if self.get_init_attr('centered', True, neurons):
            self.move(-(self.width-1)/2,-(self.height-1)/2,-(self.depth-1)/2)

    def new_iteration(self, neurons):
        return









from PymoNNto.Exploration.Network_UI.TabBase import *
import scipy.fftpack

class fourier_tab(TabBase):

    def __init__(self, parameter, title='fourier', timesteps=1000):
        super().__init__(title)
        self.parameter = parameter
        self.timesteps=timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        try:#if hasattr(neuron_group, self.parameter):
            Network_UI.add_recording_variable(neuron_group, 'np.mean(n.'+self.parameter+')', timesteps=self.timesteps)
        except:
            print(self.parameter, 'cannot be added to recorder')

    def initialize(self, Network_UI):
        self.fourier_tab = Network_UI.Next_Tab(self.title)

        curves, exc_plt = Network_UI.Add_plot_curve('Selected Group Frequencies', colors=[(0, 0, 0)], legend=False, number_of_curves=6, return_plot=True, x_label='Frequency [Hz]', y_label='Amplitude')

        self.exc_fft_curve, self.exc_alpha, self.exc_beta, self.exc_theta, self.exc_delta, self.exc_gamma = curves

        self.exc_fft_curve.setData([0, 0], [0, 0.1])

        text = pg.TextItem(text='Delta', anchor=(0.5, 0))
        text.setPos(2, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Theta', anchor=(0.5, 0))
        text.setPos(6, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Alpha', anchor=(0.5, 0))
        text.setPos(11, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Beta', anchor=(0.5, 0))
        text.setPos(22, 0)
        exc_plt.addItem(text)

        text = pg.TextItem(text='Gamma', anchor=(0, 0))
        text.setPos(30, 0)
        exc_plt.addItem(text)

        Network_UI.Next_H_Block()

        #Network_UI.Next_H_Block()

        Network_UI.Add_element(QLabel('Min:'))

        self.cut_slider = QSlider(1)
        self.cut_slider.setMinimum(0)
        self.cut_slider.setMaximum(10)
        self.cut_slider.setSliderPosition(1)
        Network_UI.Add_element(self.cut_slider)

        Network_UI.Next_H_Block()

        self.mspc_label=QLabel('ms/cycle:')
        Network_UI.Add_element(self.mspc_label)

        self.ms_per_cycle_slider = QSlider(1)
        self.ms_per_cycle_slider.setMinimum(1)
        self.ms_per_cycle_slider.setMaximum(100)
        self.ms_per_cycle_slider.setSliderPosition(10)
        Network_UI.Add_element(self.ms_per_cycle_slider)

        Network_UI.Next_H_Block()

        Network_UI.Add_element(QLabel('Smoothing:'))

        self.smooth_slider = QSlider(1)
        self.smooth_slider.setMinimum(0)
        self.smooth_slider.setMaximum(50)
        self.smooth_slider.setSliderPosition(0)
        Network_UI.Add_element(self.smooth_slider)

        #self.fouriers=[]

    def smooth(self, data):
        smoothed = data.copy()
        if len(data) > 1:
            for _ in range(int(self.smooth_slider.sliderPosition())):
                for i in range(len(data)):
                    if i == 0:
                        smoothed[i] = (smoothed[i] + smoothed[i + 1] + data[i]) / 3.0
                    elif i == len(data) - 1:
                        smoothed[i] = (smoothed[i] + smoothed[i - 1] + data[i]) / 3.0
                    else:
                        smoothed[i] = (smoothed[i] + smoothed[i - 1] + smoothed[i + 1]) / 3.0
        return smoothed

    def update(self, Network_UI):
        if self.fourier_tab.isVisible():
            if len(Network_UI.network[Network_UI.neuron_select_group]) >= 0:
                group=Network_UI.network[Network_UI.neuron_select_group, 0]

                try:
                #if hasattr(group, self.parameter):

                    cut = int(self.cut_slider.sliderPosition())
                    ms_per_cycle = int(self.ms_per_cycle_slider.sliderPosition())
                    self.mspc_label.setText('ms/cycle: {}'.format(ms_per_cycle))

                    # for name, min_f, max_f, color in [('alpha',8,13,'green'),('alpha',8,13,'green')]

                    # c2 = plt.plot([2, 1, 4, 3], pen='g', fillLevel=0, fillBrush=(255, 255, 255, 30), name='green plot')
                    # c3 = plt.addLine(y=4, pen='y')

                    #rec = group['rec_'+str(self.timesteps), 0]
                    #rec = Network_UI.rec(group, self.timesteps)
                    exc_act = group['np.mean(n.'+self.parameter+')', 0, 'np'][-self.timesteps:]
                    N = len(exc_act)
                    T = ms_per_cycle / 1000
                    yf = scipy.fftpack.fft(exc_act)
                    xf = np.linspace(int(1.0 - N / 1000), int(1.0 / (2.0 * T)), int(N / 2))[cut:]
                    real = (2.0 / N * np.abs(yf[:N // 2]))[cut:]
                    self.exc_fft_curve.setData(xf, self.smooth(real))

                    # if len(exc_act)==1000:
                    #    self.fouriers.append(real)

                    # if Network_UI.network.iteration==2000:
                    #    plt.imshow(np.array(self.fouriers).transpose())
                    #    plt.show()

                    max_y = np.max(real) + 0.001
                    max_x = np.max(xf)

                    self.exc_delta.setData([0.1, 3.9], [0, 0], fillLevel=max_y, brush=(50, 250, 50, 50), pen=None)
                    self.exc_theta.setData([min(max_x, 4), min(max_x, 7.9)], [0, 0], fillLevel=max_y, brush=(100, 200, 50, 50), pen=None)
                    self.exc_alpha.setData([min(max_x, 8), min(max_x, 13.9)], [0, 0], fillLevel=max_y, brush=(150, 150, 50, 50), pen=None)
                    self.exc_beta.setData([min(max_x, 14), min(max_x, 29.9)], [0, 0], fillLevel=max_y, brush=(200, 100, 50, 50), pen=None)
                    self.exc_gamma.setData([min(max_x, 30), min(max_x, 100)], [0, 0], fillLevel=max_y, brush=(250, 50, 50, 50), pen=None)
                except:
                    print(self.parameter, 'cannot be evaluated')

                '''
                if len(Network_UI.network[Network_UI.inh_group_name]) > 0:
                    inh_act = Network_UI.network[Network_UI.inh_group_name, 0]['np.mean(n.output)', 0, 'np'][-1000:]
                    N = len(inh_act)
                    T = ms_per_cycle / 1000
                    yf = scipy.fftpack.fft(inh_act)
                    xf = np.linspace(1.0-N/1000, 1.0 / (2.0 * T), N / 2)
                    real = (2.0 / N * np.abs(yf[:N // 2]))[cut:]
                    self.inh_fft_curve.setData(xf[cut:], self.smooth(real))

                    max_y = np.max(real)+0.001
                    max_x = np.max(xf)

                    self.inh_delta.setData([0.1, 3.9], [0, 0], fillLevel=max_y, brush=(50, 250, 50, 50), pen=None)
                    self.inh_theta.setData([min(max_x, 4), min(max_x, 7.9)], [0, 0], fillLevel=max_y, brush=(100, 200, 50, 50), pen=None)
                    self.inh_alpha.setData([min(max_x, 8), min(max_x, 13.9)], [0, 0], fillLevel=max_y, brush=(150, 150, 50, 50), pen=None)
                    self.inh_beta.setData([min(max_x, 14), min(max_x, 29.9)], [0, 0], fillLevel=max_y, brush=(200, 100, 50, 50), pen=None)
                    self.inh_gamma.setData([min(max_x, 30), min(max_x, 100)], [0, 0], fillLevel=max_y, brush=(250, 50, 50, 50), pen=None)
                '''




import sys
import os
#import Exploration.StorageManager.StorageManager as sm
sys.path.append('../../')


def args_to_param_dict(add=[]):
    result = {}
    for arg in add+sys.argv[1:]:
        split = arg.split('=')
        if len(split) == 2:
            key, value  =split
            try:
                result[key] = eval(value)
            except:
                result[key] = value
        else:
            print('unknown parameter', arg)
    return result

params=args_to_param_dict()


import_file=params.get('import_file', 'Testing.SORN_Grammar.GrammarExperiment_Hierarchical')
exec('from ' + import_file + ' import *')

func_name=params.get('func_name', 'run')
evaluation_function = eval(func_name)

score = evaluation_function(params)

print('score:', score)

#name = sys.argv[1]
#individual = sys.argv[2]
#import_file = sys.argv[3]
#func_name = sys.argv[4]

#individual=eval(individual)
#exec('from ' + import_file + ' import *')
#evaluation_function = eval(func_name)

#params = {'evolution': True}
#if len(sys.argv) > 5:
#    for arg in sys.argv[5:]:
#        if '=' in arg:
#            key, value = arg.split('=')
#            params[key] = eval(value)
#        else:
#            print('no = in argument! (ignored)')
#else:
#    params['N_e'] = 900
#    params['TS'] = [1]

#sm.get_data_folder()+

#folder = "../../Data/Evo"

#if not os.path.exists(folder):
#    try:
#        os.mkdir(folder)
#    except:
#        print("main folder already exists")

#text_file = open(folder+'/'+name+".txt", "w")
#n = text_file.write("score: "+str(score))
#text_file.close()



import numpy as np

from numpy.random import *


def_dtype = np.float64

class NetworkObjectBase:

    def __init__(self, tag):
        self.tags = []
        self.clear_cache()
        if tag is not None:
            self.add_tag(tag)

    def find_objects(self, key):#override for deeper search
        return []

    def clear_cache(self):
        self.tag_shortcuts = {}

    def set_tag_attrs(self, tag, attr, value):
        for obj in self[tag]:
            setattr(obj, attr, value)

    def call_tag_functions(self, tag, attr, **args):#todo: test
        for obj in self[tag]:
            getattr(obj, attr)(args)

    def __getitem__(self, key):

        np_array_conversion = isinstance(key, tuple) and 'np' in key

        if key in self.tag_shortcuts and type(self.tag_shortcuts[key]) is not np.ndarray:
            if np_array_conversion:
                return np.array(self.tag_shortcuts[key])
            else:
                return self.tag_shortcuts[key]

        if isinstance(key, tuple):
            k = key[0]
            index = key[1]
        else:
            k = key
            index = None

        result = []
        if k in self.tags or (k is type and isinstance(self, k)):
            result = [self]

        result += self.find_objects(k)

        if index is not None:
            if len(result)>0:
                result = result[index]
            else:
                result=None

        if not key in self.tag_shortcuts:
            self.tag_shortcuts[key] = result

        if np_array_conversion:# only used for [key,index,'np'] not for [key, 'np']
            return np.array(result)
        else:
            return result


    def add_tag(self, tag):
        for subtag in tag.split(','):
            self.tags.append(subtag)
        return self

    def buffer_roll(self, mat, new=None):
        #return np.roll(mat, 1, axis=0)
        mat[1:len(mat)] = mat[0:len(mat) - 1]

        if new is not None:
            mat[0]=new

        return mat

    def get_nparray(self, dim):
        return np.zeros(dim).astype(def_dtype)

    def get_random_nparray(self, dim, density=None, clone_along_first_axis=False, rnd_code=None):#rnd_code=random_sample(dim)
        if rnd_code is None:
            result = random_sample(dim)
        else:
            if 'dim' not in rnd_code:
                if rnd_code[-1] == ')':
                    rnd_code = rnd_code[:-1]+',size=dim)'
                else:
                    rnd_code = rnd_code+'(size=dim)'
            result = eval(rnd_code)

        if density is None:
            result = result.astype(def_dtype)
        elif type(density) == int or type(density) == float:
            result = (result * (random_sample(dim) <= density)).astype(def_dtype)
        elif type(density) is np.ndarray:
            result = (result * (random_sample(dim) <= density[:, None])).astype(def_dtype)


        if not clone_along_first_axis:
            return result
        else:
            return np.array([result[0] for _ in range(dim[0])])


    def get_buffer_mat(self, dim, size):
        return np.array([self.get_nparray(dim) for _ in range(size)])



from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *
import random
import numpy as np
from collections import Counter
import pypianoroll as piano
import os
import re
from scipy.stats import entropy
from sklearn.feature_extraction.text import CountVectorizer



#{' PERIOD': '.', ' COMMA': '.', ' QUESTION': '.'} #, 'MNAME':'Name', 'FNAME':'Name'
#CHILDES_activator = TextActivator_New(filename='CHILDES.txt', replace_dict={' PERIOD': '.', ' COMMA': '.', ' QUESTION': '.'})
#SPD_activator = TextActivator_New(filename='SPD.txt', replace_dict={'?': '.', ':': '.', ',': '.', '.': '. ', '  ': ' ', '-': ''})


class MusicActivator(PatternGroup):

    def initialize(self): #
        self.last_iteration=-1
        self.nextone=None

        self.notes = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B'] # length of octave
        self.octaves = list(range(11))
        self.is_drum = False
        self.max_iterations = self.kwargs.get('max_iterations', None)

        self.beat_resolution = self.kwargs.get('beat_resolution', 24)
        self.instrument = self.kwargs.get('instrument', int(1))
        self.tempo = self.kwargs.get('tempo', int(120)) # BPM

        self.which_alphabet = self.kwargs.get('which_alphabet', 'train') 
        assert self.which_alphabet in ['train', 'minmax', 'all']
        self.current_index = -1

        self.seen = [] # part of corpus that has already been seen
        self.times_seen = 0 # times we have iterated through whole corpus
        
    def midi_index_to_notestring(self, index):
        '''
        Convert MIDI index to a string of the note

        Arguments:
        index -- int, MIDI index from 0-127 or -1 for silence

        Returns:
        string -- string corresponding to a note
        '''
        if index == -1:
            note_as_string = '_'
        else:
            octave = (index // len(self.notes))-1
            assert octave in self.octaves, 'Illegal MIDI index'
            assert 0 <= index <= 127, 'Illegal MIDI index'
            note = self.notes[index % len(self.notes)]

            note_as_string = note + str(octave)

        return note_as_string

    def load_MIDI(self, path):
        multitrack = piano.parse(path)
        multitrack.trim_trailing_silence()
        multitrack.binarize()
        single_tracks = np.empty((0, 128))

        if self.type=='mono':
            for i in range(len(multitrack.tracks)):
                singletrack = multitrack.tracks[i].pianoroll

                if multitrack.beat_resolution != self.beat_resolution: # we downsample
                    # check if the shortest note is not shorter than our beat resolution
                    count = 0
                    legal = True
                    for i in range(len(singletrack)-1):
                        if np.any(singletrack[i]) and not np.any(singletrack[i+1]) and count < int(multitrack.beat_resolution/self.beat_resolution):
                            legal = False
                            break
                        if np.any(singletrack[i]) and np.array_equal(singletrack[i],singletrack[i+1]) :
                            count+=1
                        else:
                            count = 0

                    if legal: # shortest pitch in track is representable by our beat resolution
                        downsampled_track = []
                        downsampled_track.append(singletrack[0]) # first time step always in
                        for ts in range(1,len(singletrack)-1):
                            if ts%int(multitrack.beat_resolution/self.beat_resolution)==0 or not np.any(singletrack[ts]) and np.any(singletrack[ts-1]) and np.any(singletrack[ts+1]):
                                downsampled_track.append(singletrack[ts])
                        if (len(singletrack)-1)%int(multitrack.beat_resolution/self.beat_resolution) ==0 :
                            downsampled_track.append(singletrack[len(singletrack)-1]) # also consider last time step 

                        single_tracks = np.append(single_tracks, np.array(downsampled_track), axis=0)
                    else:
                        print('skip track that is not representable by desired beat resolution', self.beat_resolution)


                elif self.beat_resolution == multitrack.beat_resolution:
                    # we don't downsample because beat resolution is the same as the track
                    single_tracks = np.append(single_tracks, np.array(singletrack), axis=0)

        if self.type=='poly':
            # merge the tracks of the same instrument
            instruments= {}
            instrument = -1
            for i in range(len(multitrack.tracks)):
                if instrument != multitrack.tracks[i].program:
                    instruments[multitrack.tracks[i].program] = [i] # append track IDs to instrument
                    instrument = multitrack.tracks[i].program
                elif instrument == multitrack.tracks[i].program:
                    instruments[instrument].append(i)

            merged_tracks_per_instrument = []

            for instrument in instruments:
                merged_tracks_per_instrument.append(multitrack[instruments[instrument]].get_merged_pianoroll())

            if multitrack.beat_resolution != self.beat_resolution:
                # we want to find out if we can downsample the track
                # we need to know the shortest and longest note
                merged = multitrack.get_merged_pianoroll() # all tracks merged together

                # make a list of all possible notes 
                indices_non_zero = np.where(merged)
                alphabet = list(np.arange(21,109))
                indices_per_note = {}
                min_per_note = {}
                max_per_note = {}            

                # get indices in track where each note is played
                for note in alphabet:
                    indices_per_note[note] = indices_non_zero[0][np.where(indices_non_zero[1]==note)]
                    if indices_per_note[note].size > 0: # else the note does not appear (length 0)
                        min_per_note[note] = len(indices_per_note)
                        max_per_note[note] = 1
                        count = 1
                        for i in range(len(indices_per_note[note])-1):
                            if indices_per_note[note][i+1] == indices_per_note[note][i]+1:
                                count+=1
                            else:
                                if count < min_per_note[note]:
                                    min_per_note[note] = count
                                if count > max_per_note[note]:
                                    max_per_note[note] = count
                                count = 1

                shortest_note = int(min(min_per_note.values()))
                longest_note = int(max(max_per_note.values()))

                # get shortest possible beat resolution
                #for i in reversed(range(shortest_note)):
                #    if beat_resolution%(i+1) == 0:
                #        beat_resolution = int(beat_resolution/(i+1))
                #        break       

                # is shortest note a divisor of beat_resolution?
                if shortest_note >= int(multitrack.beat_resolution/self.beat_resolution):
                #self.beat_resolution%shortest_note ==0 and shortest_note!=1:
                    # then we down-sample
                    for track in merged_tracks_per_instrument:
                        downsampled_track = []
                        downsampled_track.append(track[0]) # first time step always in
                        for ts in range(1,len(track)-1):
                            if ts%int(multitrack.beat_resolution/self.beat_resolution)==0 or not np.any(track[ts]) and np.any(track[ts-1]) and np.any(track[ts+1]):
                                downsampled_track.append(track[ts])
                        if (len(track)-1)%int(multitrack.beat_resolution/self.beat_resolution)==0:
                            downsampled_track.append(track[len(track)-1]) # also consider last time step 

                        single_tracks = np.append(single_tracks, np.array(downsampled_track), axis=0)


            if multitrack.beat_resolution == self.beat_resolution:
                for track in merged_tracks_per_instrument:
                    single_tracks = np.append(single_tracks, np.array(track), axis=0)
            
            # else we ignore the track because it is not representable by the chosen beat resolution
        return single_tracks

    def get_alphabet_length(self):
        return len(self.alphabet)

    def generate_weights(self, neurons):

        if hasattr(neurons, 'input_density'):
            density = neurons.input_density
        else:
            density = self.kwargs.get('input_density', 1/60)#int(output_size / 60)

        output_size = neurons.size
        #output_size = self.kwargs.get('output_size', 600)
        # self.kwargs.get('activation_size', int(output_size / 60))

        A = self.get_alphabet_length()

        self.activation_size = int(output_size * density)

        neurons.Input_Weights = np.zeros((output_size, A))
        available = set(range(output_size))
        for a in range(A):
            temp = random.sample(available, self.activation_size)
            neurons.Input_Weights[temp, a] = 1
            available = set([n for n in available if n not in temp])
            assert len(available) > 0, 'Input alphabet too big for non-overlapping neurons'

        neurons.Input_Mask = (np.sum(neurons.Input_Weights, axis=1) > 0)
        neurons.mean_network_input_activity=self.activation_size/output_size

    def get_pattern(self, neurons): 
        if self.active:
            n_hot = self.get_next(next=neurons.iteration != self.last_iteration)
            self.last_iteration=neurons.iteration
            self.current_pattern_index = n_hot #for pattern activator   --> how does this work if my pattern is more than one symbol at a time?
            return self.get_activation(n_hot, neurons) #self.W[:, i]
        else:
            return None

    def initialize_neuron_group(self, neurons):
        #if not hasattr(neurons, 'Input_Weights'):
        self.generate_weights(neurons)

    def get_dominant_key(self, pianoroll):
        assert pianoroll.shape[1] == 128
        kind = ['minor', 'major']
        max_perc = 0
        key = np.zeros(3) # base tone, minor(0)/major(1), percent
        for i in range(12):
            for j in range(2):
                curr_perc = piano.metrics.in_scale_rate(pianoroll, key=i, kind=kind[j])
                if curr_perc > max_perc:
                    max_perc = curr_perc
                    key[0] = i
                    key[1] = j
                    key[2] = curr_perc
        return {'key': [key[0], kind[int(key[1])]], 'percentage': key[2]}

    def albphabet_index_to_note_symbol(self, index):
        """
        Convert an alphabet index (from the network) to the corresponding note string

        Arguments:
        index -- int, alphabet index

        Returns:
        symbol -- string corresponding to note
        """

        # first get the MIDI index
        midi = self.alphabet[index]
        return self.midi_index_to_notestring(midi)


    def midi_index_to_alphabet_index(self, midi):
        '''
        looks for the MIDI index in the alphabet of the SORN and returns its index in the alphabet

        Arguments:
        midi -- int, a MIDI index

        Returns:
        index -- int, an index to index the networks alphabet
        '''
        return self.alphabet.index(midi)



class PolyphonicActivator(MusicActivator):

    def initialize(self):
        super().initialize()
        self.type = 'poly'
        self.path_to_music = self.kwargs.get('path_to_music', os.path.join(os.path.dirname(__file__),'..', '..','..','Data/midis/polyphonic'))
        self.input_midis = self.get_MIDIpianoroll() # sequence of vectors with shape (128)
        self.alphabet = self.get_alphabet()
        self.corpus = self.get_corpus()
        self.A = len(self.alphabet)
        self.lowest_pitch = min(self.alphabet)
        self.highest_pitch = max(self.alphabet) 
        self.dominant_key = self.get_dominant_key(self.convert_sequence_to_pianoroll(self.corpus))
        self.max_iterations = len(self.corpus)

    def get_note_input_statistics_list(self):
        return [len(np.nonzero(self.corpus[:,i])[0]) for i in range(len(self.alphabet))]

    def get_MIDIpianoroll(self):
        # creates polyphonic pianoroll of length max iterations
        corpus = np.empty((0, 128)).astype(bool)
        for i in os.listdir(self.path_to_music):
            if i.endswith('.mid') or i.endswith('.midi'):
                corpus = np.append(corpus, self.load_MIDI(os.path.join(self.path_to_music, i)), axis=0)
                if self.max_iterations is not None and len(corpus) > self.max_iterations:
                    corpus = corpus[:self.max_iterations]
                    # stop reading in MIDIs if max corpus size is reached
                    break
        assert np.any(corpus), 'empty corpus'
        return corpus

    def get_alphabet(self):
        if self.which_alphabet == 'minmax':
            indices_non_zero = self.input_midis.nonzero()
            alphabet = list(range(indices_non_zero[1].min(), indices_non_zero[1].max()))
            return alphabet
        elif self.which_alphabet == 'train':
            indices_non_zero = self.input_midis.nonzero()
            alphabet = sorted(set(indices_non_zero[1])) 
            return alphabet         
        elif self.which_alphabet == 'all': # use all notes on a grand piano
            return list(range(21,109)) 

    def get_corpus(self):
        return self.input_midis[:, self.alphabet]

    def convert_sequence_to_pianoroll(self, sequence):
            'conerts a sequence of MIDI indices to a pianoroll'
            pianoroll = np.zeros((len(sequence), 128)).astype(np.bool_)
            pianoroll[:, self.alphabet] = sequence
            return pianoroll

    '''
    def get_dominant_key(self, sequence=None):
        # get dominant key in corpus 
        pianoroll = self.convert_sequence_to_pianoroll(sequence)
        return super.get_dominant_key(pianoroll)
    '''
    def set_nextone(self, vec):
        'sets next vector'
        self.nextone = vec


    def get_next(self, next=False):
        if next:
            self.current_index += 1

        # in case the sequence ends, restart it
        if self.current_index == len(self.corpus):
            self.current_index = 0
            self.times_seen+=1

        if self.nextone is not None:#outside activation
            self.corpus[self.current_index,:] = self.nextone

            self.nextone=None
         
        if next:
            self.seen.append(self.corpus[self.current_index,:])
            # corpus is already sequence of n-hot arrays, we do not need to convert it
        return self.corpus[self.current_index,:].astype(bool)


    def get_activation(self, n_hot, neurons):
        if not np.any(n_hot):
            return np.zeros(neurons.Input_Weights.shape[0]).astype(bool) # return a vector of zeros
        else:
            vector = np.zeros(neurons.Input_Weights.shape[0])
            indices = np.where(n_hot == 1)[0]
            for i in indices:
                vector += neurons.Input_Weights[:,i]
            return vector.astype(bool)
            #return neurons.Input_Weights[:,n_hot]
            #return neurons.Input_Weights * mask



    def get_music_score(self, spont_output, pianoroll, seen=None):
        result = {}

        output = spont_output[:-1]
        output=output.split(' ')

        if seen is None: # no network output sequence was given
            seen = self.seen[:len(self.seen)-len(output)] # the last n time steps it gets fed in the previous spontaneous output: that's not comparable between experiments

        seen_string = ''
        for i in range(len(seen)-len(output)): # the last n time steps it gets fed in the previous spontaneous output: that's not comparable between experiments
            if np.any(seen[i]): # vector is not all zeros
                indices = np.nonzero(seen[i])[0]
                if len(indices)==1: # predict only one note at this time step
                    note = self.albphabet_index_to_note_symbol(indices[0])
                    seen_string += note+' '
                else:
                    for j in indices:
                        note = self.albphabet_index_to_note_symbol(j)
                        seen_string += note
                    seen_string += ' '
            else:
                seen_string += '_ ' # silence

        seen_string = seen_string.split(' ')
        seen_string = seen_string[:-1]

        result['seen_string']=seen_string
        result['times_seen'] = self.times_seen
        result['spont_output'] = spont_output
        result['output_pianoroll'] = pianoroll
        result['seen_pianoroll'] = seen

        result['total_score']=0

        # compare frequencies of notes 

        # make a list of all possible notes 
        indices_non_zero = np.where(pianoroll)
        number_note_played = {}
        number_note_seen = {}

        indices_non_zero_seen = np.where(seen)

        # NEED TO MAKE SURE TO COMPARE STRINGS WITH STRINGS OR NUMBERS WITH NUBMERS HERE
        for note in set(indices_non_zero_seen[1]): # can't take alphabet bc perhaps it has not seen all of corpus
            number_note_played[self.albphabet_index_to_note_symbol(note)] = len(indices_non_zero[0][np.where(indices_non_zero[1]==self.alphabet[note])])
            number_note_seen[self.albphabet_index_to_note_symbol(note)] = len(indices_non_zero_seen[0][np.where(indices_non_zero_seen[1]==note)])

        number_note_seen['_'] = len(set(range(len(seen))) - set(indices_non_zero_seen[0]))
        number_note_played['_'] = len(set(range(len(pianoroll))) - set(indices_non_zero[0]))

        sorted_notes = [x[0] for x in sorted(number_note_seen.items(), key=lambda kv: kv[1], reverse=True)]
        freq_input = np.array([number_note_seen[x] for x in sorted_notes])
        freq_input = freq_input/freq_input.sum()

        result['input_frequencies'] = inp = dict(zip(sorted_notes, freq_input))
        result['sorted_symbols'] = sorted_notes

        freq_output = np.array([number_note_played[x] for x in sorted_notes])
        freq_output = freq_output/freq_output.sum()

        result['output_frequencies'] = out = dict(zip(sorted_notes, freq_output))

        result['diff_frequencies'] = diff = {x: abs(inp[x] - out[x]) for x in inp if x in out}

        result['sum_differences'] = sum_diff = sum(diff.values())

        result['total_score'] -= sum_diff

        # Now we look at the frequencies of chords

        times_seen_chords = {}
        notes_in_chords = {}
        for ts in range(len(seen_string)):
            if len(np.where(seen[ts])[0])>1: # only a chord if more than one note is played at the same time
                chord = seen_string[ts]
                if chord in times_seen_chords:
                    times_seen_chords[chord] +=1
                else:
                    notes_in_chords[chord] = np.where(seen[ts])[0] # how many notes are in this chord?
                    times_seen_chords[chord] = 1
          
        times_output_chords = {}

        for ts in range(len(pianoroll)):
            if len(np.where(pianoroll[ts])[0])>1:
                chord = output[ts]
                if chord in times_output_chords:
                    times_output_chords[chord] += 1
                else:
                    times_output_chords[chord] = 1
                if chord not in notes_in_chords:
                    notes_in_chords[chord] = np.where(pianoroll[ts][0])

        sorted_chords = [x[0] for x in sorted(times_seen_chords.items(), key=lambda kv: kv[1], reverse=True)]
        freq_input = np.array([times_seen_chords[x] for x in sorted_chords])
        freq_input = freq_input/freq_input.sum()

        result['input_frequencies_chords'] = inp = dict(zip(sorted_chords, freq_input))
        result['sorted_chords'] = sorted_chords

        freq_output = np.zeros(len(sorted_chords))
        for i, chord in enumerate(sorted_chords):
            if chord in times_output_chords:
                freq_output[i] = times_output_chords[chord]

        #freq_output = np.array([times_output_chords[x] for x in sorted_chords])
        freq_output = freq_output/freq_output.sum()

        result['output_frequencies_chords'] = out = dict(zip(sorted_chords, freq_output))
        result['midi_indices_per_chord'] = notes_in_chords
        result['n_notes_per_chord'] = len(notes_in_chords)

        # differences in frequencies
        # punish less if it gets frequencies of chords (compared to notes overall)
        result['diff_frequencies_chords'] = diff = {x: abs(inp[x] - out[x]) for x in inp if x in out}

        sum_diff = 0
        
        for key in diff: # punish less if it gets frequencies of higher n_grams wrong
            sum_diff += diff[key]/len(notes_in_chords[key]) # weighted by number of notes in a chord
        result['sum_differences_chords'] = sum_diff

        result['total_score'] -= sum_diff 

        #output frequencies for all output chords (also ones that do not appear in input)
        sorted_chords = [x[0] for x in sorted(times_output_chords.items(), key=lambda kv: kv[1], reverse=True)]
        freq_output = np.array([times_output_chords[x] for x in sorted_chords])
        freq_output = freq_output/freq_output.sum()

        result['sorted_chords_output_all'] = sorted_notes
        result['output_frequencies_chords_all'] = freq_output

        return result

    def get_entropy_of_track(self, pianoroll):
        alphabet = self.get_alphabet()
        indices_non_zero = pianoroll.nonzero()
        distribution = np.zeros(len(alphabet))
        for i in range(len(alphabet)):
            distribution[i] = sum(indices_non_zero[1] == alphabet[i])
        # add the silent time steps
        non_zero_indices = set(indices_non_zero[0])
        all_indices = set(range(len(pianoroll)))
        silent_indices = all_indices - non_zero_indices
        distribution=np.append(distribution, len(silent_indices))
    
        return entropy(distribution)

class MonophonicActivator(MusicActivator): # convert Pianoroll from MIDI to monophonic sequence of pitches

    def initialize(self):
        super().initialize()
        self.type = 'mono'
        self.path_to_music = self.kwargs.get('path_to_music', os.path.join(os.path.dirname(__file__),'..', '..','..','Data/midis/monophonic'))
        self.input_midis = self.get_MIDIpianoroll()
        self.corpus = self.get_corpus()
        self.alphabet = self.get_alphabet()
        self.A = len(self.alphabet)
        self.lowest_pitch = min(self.alphabet)
        self.highest_pitch = max(self.alphabet)   
        print("Alphabet:", self.alphabet)
        self.dominant_key = self.get_dominant_key(self.convert_sequence_to_pianoroll(self.corpus))
        self.max_iterations = len(self.corpus)
        print(self.get_entropy_of_track(self.corpus))

    def get_MIDIpianoroll(self):
        # creates polyphonic pianoroll of length max iterations
        full_roll = np.empty((0, 128)).astype(bool)
        for i in os.listdir(self.path_to_music):
            if i.endswith('.mid') or i.endswith('.midi'):
                p_roll = self.load_MIDI(os.path.join(self.path_to_music, i))
                if np.any(p_roll) and piano.metrics.polyphonic_rate(p_roll,threshold=1) > 0: # we only work with monophonic input when we have symbolic alphabet
                    print("skip polyphonic track")
                    pass
                else:
                    full_roll = np.append(full_roll, p_roll, axis=0)
                    if self.max_iterations is not None and len(full_roll) > self.max_iterations:
                        full_roll = full_roll[:self.max_iterations]
                        # stop reading in MIDIs if max corpus size is reached
                        break
        assert np.any(full_roll), 'empty corpus'
        return full_roll                        

    def get_alphabet(self):
        if self.which_alphabet == 'train': # only the notes that also appear in training data
            return sorted(set(self.corpus))

        elif self.which_alphabet == 'minmax': # all notes between lowest and highest pitch in training data
            min_pitch = self.corpus[np.where(self.corpus > 0, self.corpus, np.inf).argmin()] # need to ignore -1
            max_pitch = self.corpus.max()
            alphabet = list(range(min_pitch, max_pitch+1))
            alphabet.append(-1) # add this for silence
            return alphabet

        elif self.which_alphabet == 'all': # all possible notes on a grand piano (MIDI IDs 21-108)
            alphabet = list(range(21,109))
            alphabet.append(-1) # add this for silence (all zeros)
            return alphabet

    def get_corpus(self):
        self.input_midis = self.input_midis[:self.max_iterations]
        sym_sequence = np.full(self.input_midis.shape[0], -1).astype(np.int8) # make a sequence of -1 (stands for silence)
        indices_non_zero = self.input_midis.nonzero() # find all time steps where a tone is played
        sym_sequence[indices_non_zero[0]] = indices_non_zero[1] # set non_zero time steps with the current index for pitch played
        corpus = np.array(sym_sequence).flatten().astype(np.int8)
        return corpus

    def get_note_input_statistics_list(self):
        result = np.zeros(self.get_alphabet_length())
        for ts in range(len(self.corpus)):
            for index in range(len(self.alphabet)):
                if self.corpus[ts] == self.alphabet[index]:
                    result[index] +=1
        return result

    def set_nextone(self, midi):
        'sets next index'
        self.nextone = midi

    def get_next(self, next=False):

        if next:
            self.current_index += 1

        if self.max_iterations is not None and self.current_index % self.max_iterations == 0:
            self.current_index = 0
            self.times_seen+=1

        if self.nextone is not None:#outside activation (spontaneous phase: we feed output back in)
            self.corpus[self.current_index] = self.nextone
            self.nextone=None

        if next:
            self.seen.append(self.corpus[self.current_index])

        return self.corpus[self.current_index]

    def print_test(self, length=1000):
        temp = self.current_index
        t = ''
        for i in range(length):
            c = self.corpus[i]
            t += self.index_to_symbol(c)
        print(t)
        self.current_index = temp
        return self

    def convert_sequence_to_pianoroll(self, sequence):
        '''
        converts sequence of MIDI indices to a pianoroll
        '''
        pianoroll = np.empty((len(sequence), 128)).astype(bool)
        for i in range(len(sequence)):
            index = sequence[i]
            one_hot = np.zeros(128) # make one-hot vector
            if index != -1:
                one_hot[index] = True
            pianoroll[i] = one_hot
        return pianoroll

    '''
    def get_dominant_key(self, pianoroll=None):
        if pianoroll == None:
            pianoroll = self.corpus
        proll = self.convert_sequence_to_pianoroll(pianoroll)
        return super.get_dominant_key(pianoroll=proll)
    '''
    def initialize_neuron_group(self, neurons):
        #if not hasattr(neurons, 'Input_Weights'):
        self.generate_weights(neurons)

    def get_activation(self, pitch_index, neurons):
        return neurons.Input_Weights[:, pitch_index]

    def get_pattern(self, neurons):
        if self.active:
            c = self.get_next(next=neurons.iteration != self.last_iteration)
            self.last_iteration=neurons.iteration
            i = self.midi_index_to_alphabet_index(c)
            self.current_pattern_index = i #for pattern activator
            return self.get_activation(i, neurons) #self.W[:, i]
        else:
            return None

    def get_entropy_of_track(self, sequence):
        alphabet = self.get_alphabet()
        distribution = np.zeros(len(alphabet))
        for i in range(len(alphabet)):
            distribution[i] = len(np.where(sequence ==alphabet[i])[0])
        return entropy(distribution)

    def get_music_score(self, spont_output, pianoroll, seen=None):
        result={}

        output = spont_output[:-1]
        output=output.split(' ')

        if seen is None: # no MIDI sequence was given
            seen = self.seen
        
        seen_string = []
        for i in range(len(seen)-len(output)): # the last n time steps it gets fed in the previous spontaneous output: that's not comparable between experiments
            seen_string.append(self.midi_index_to_notestring(seen[i]))

        #result['seen']=seen_string
        result['times_seen'] = self.times_seen
        result['spont_output'] = spont_output

        # n-gram frequencies
        result['total_score']=0 # this is the best possible score if there are no differences in frequencies

        for i in range(1, 4):
            input_igrams = [seen_string[j:j+i] for j in range(len(seen_string)-i+1)]  
            input_igrams = [' '.join([elem for elem in input_igrams[x]]) for x in range(len(input_igrams))]
            notes_input = Counter(input_igrams)
            sorted_notes = [x[0] for x in sorted(notes_input.items(), key=lambda kv: kv[1], reverse=True)]
            freq_input = np.array([notes_input[x] for x in sorted_notes])
            freq_input = freq_input/freq_input.sum()
            result['input_frequencies_{}'.format(i)] = inp = dict(zip(sorted_notes, freq_input))     
            result['sorted_symbols_{}'.format(i)] = sorted_notes

            output_igrams = [output[j:j+i] for j in range(len(output))]
            output_igrams = [' '.join([elem for elem in output_igrams[x]]) for x in range(len(output_igrams))]
            output_counter = Counter(output_igrams)
            freq_output = np.array([output_counter[x] for x in sorted_notes])
            freq_output = freq_output/freq_output.sum()
            result['output_frequencies_{}'.format(i)] = out = dict(zip(sorted_notes, freq_output))

            result['diff_frequencies_{}'.format(i)] = diff = {x: abs(inp[x] - out[x]) for x in inp if x in out}

            result['sum_differences_{}'.format(i)] = sum_diff = sum(diff.values())

            result['total_score'] -= sum_diff/i # punish less if it gets frequencies of higher n_grams wrong

        #store pitch transitions
        result['n_different_transitions'] = 0
        for i in range(len(spont_output) - 1):
            if spont_output[i] != spont_output[i + 1]:
                result['n_different_transitions'] += 1

        if (len(spont_output) - 1)>0:
            result['%_different_transitions'] = result['n_different_transitions']/(len(spont_output) - 1)
        else:
            result['%_different_transitions'] = 0.0
        
        result['dominant_key_output'] = self.get_dominant_key(pianoroll)

        result['dominant_key_input'] = self.get_dominant_key(self.convert_sequence_to_pianoroll(seen))
    
        return result


from PymoNNto.NetworkBehaviour.Input.Pattern_Basics import *

from PIL import Image, ImageDraw, ImageFilter, ImageEnhance

from PymoNNto.NetworkBehaviour.Input.Images.Helper import *

from PymoNNto.NetworkBehaviour.Input.Images.PixelPattern import *

#from scipy.ndimage import *
#from scipy import signal
#import scipy.ndimage.filters as fi
#import matplotlib.pyplot as plt

class TNAP_Image_Patches(Pixel_Pattern):

    def dim_contains(self, str):
        for dim in self.dimensions:
            for s in str:
                if s in dim:
                    return True
        return False


    def get_filters(self):

        print('rendering images...')

        self.image_path = self.kwargs.get('image_path', '')
        self.dimensions = self.kwargs.get('dimensions', ['ImageFilter.FIND_EDGES'])
        self.channel_norm = self.kwargs.get('channel_norm', 'max')

        pil_image = Image.open(self.image_path)
        pil_image_gray = pil_image.convert('L')
        image_rgb = np.array(pil_image).astype(np.float64)
        white = np.array(pil_image_gray).astype(np.float64)

        if self.dim_contains(['red']): red = image_rgb[:, :, 0]
        if self.dim_contains(['green']): green = image_rgb[:, :, 1]
        if self.dim_contains(['blue']): blue = image_rgb[:, :, 2]
        if self.dim_contains(['rgbw']): rgbw = get_rgbw(image_rgb[:, :, 0], image_rgb[:, :, 1], image_rgb[:, :, 2], white)
        #if self.dim_contains(['center_red']): on_center_red, off_center_red = get_LOG_On_Off(image_rgb[:, :, 0])
        #if self.dim_contains(['center_green']): on_center_green, off_center_green = get_LOG_On_Off(image_rgb[:, :, 1])
        #if self.dim_contains(['center_blue']): on_center_blue, off_center_blue = get_LOG_On_Off(image_rgb[:, :, 2])
        if self.dim_contains(['center_white']): on_center_white, off_center_white = get_LOG_On_Off(white)
        if self.dim_contains(['center_red_green']): on_center_red_green, off_center_red_green = get_multi_color_LOG_On_Off(image_rgb[:, :, 0], image_rgb[:, :, 1])
        if self.dim_contains(['center_yellow_blue']): on_center_yellow_blue, off_center_yellow_blue = get_multi_color_LOG_On_Off(image_rgb[:, :, 0]*image_rgb[:, :, 1], image_rgb[:, :, 2])
        if self.dim_contains(['center_green_red']): on_center_green_red, off_center_green_red = get_multi_color_LOG_On_Off(image_rgb[:, :, 1], image_rgb[:, :, 0])
        if self.dim_contains(['center_blue_yellow']): on_center_blue_yellow, off_center_blue_yellow = get_multi_color_LOG_On_Off(image_rgb[:, :, 2], image_rgb[:, :, 0]*image_rgb[:, :, 1])


        filters = []
        for dim in self.dimensions:

            res = eval(dim)

            if res is object and issubclass(res, ImageFilter.BuiltinFilter):
                res = np.array(pil_image_gray.filter(res))

            #print(dim)
            #plt.imshow(res, cmap=plt.cm.gist_gray, interpolation='nearest')#filters[-1]
            #plt.show()

            if self.channel_norm == 'max':
                filters.append(res / np.maximum(np.max(res), 1))
            elif self.channel_norm == 'sum':
                filters.append(res / np.maximum(np.sum(res), 1))
            else:
                filters.append(res)



        return np.array(filters)



    def initialize(self):
        super().initialize()
        self.patch_norm = self.kwargs.get('patch_norm', True)

        self.patterns.append(self.get_filters())
        self.labels.append('image_patch')

        self.grid_channels = len(self.patterns[0])


    #def get_image_patch(self):
    #    x = np.random.randint(self.dim_matrices.shape[2] - self.grid_width)
    #    y = np.random.randint(self.dim_matrices.shape[1] - self.grid_height)
    #    submat = self.dim_matrices[:, y:y + self.grid_height, x:x + self.grid_width].flatten()

    #    return submat


    def get_pattern_values(self, pattern):
        i = 0
        result = np.zeros(1)
        while i < 10 and np.sum(result) < 1.0:
            i += 1
            result = super().get_pattern_values(pattern)

        if self.patch_norm:
            d = np.max(result)
            if d == 0:
                d = 1
            return result/d
        else:
            return result

    #def get_pattern(self):
    #    i = 0
    #    submat = np.zeros(1)
    #    while i < 10 and sum(submat) < 1.0:
    #        i += 1
    #        submat = self.get_image_patch()

    #    if self.patch_norm:
    #        d = np.max(submat)
    #        if d == 0:
    #            d = 1
    #        return submat/d
    #    else:
    #        return submat


    #def get_pattern_dimension(self, remove_depth=True):
    #    if remove_depth:
    #        return [self.grid_height * self.dim_matrices.shape[0], self.grid_width]
    #    else:
    #        return [self.dim_matrices.shape[0], self.grid_height, self.grid_width]

    #def reconstruct_pattern(self, p):
    #    return p.reshape(self.grid_height * self.dim_matrices.shape[0], self.grid_width)






class TNAP_Foveation_Image_Patches(TNAP_Image_Patches):

    def get_multi_dim_resort_mask(self, resort_mask, depth):
        result = []
        print(resort_mask.shape)
        for i in range(depth):
            result.append(resort_mask+(self.hup*self.wup)*i)#resort_mask.shape[0]
            #result = np.concatenate([result, resort_mask+resort_mask.shape[0]*i])
        return np.array(result).flatten().astype(np.int32)

    def get_receptive_field_resort_mask(self, w, h, tw, th, sharpness, depth):
        result_img = np.zeros((w, h))

        for x in range(w):
            for y in range(h):
                dx = w / 2 - x
                dy = h / 2 - y
                dist = np.sqrt(dx * dx + dy * dy)
                d = np.clip(1 - (dist) / (w / 2), 0, None)
                result_img[x, y] = np.maximum(np.power(d, sharpness), 0.0000001)

        rif = result_img.flatten()
        rif = rif / sum(rif)
        result = np.random.choice(np.array(list(range(rif.shape[0]))), tw * th, False, rif)

        return result_img, self.get_multi_dim_resort_mask(result, depth)

    def compress_receptive_field(self, patch, resort_mask):
        return patch.flatten()[resort_mask].reshape(self.grid_channels, self.grid_height, self.grid_width)#self.

    def decompress_receptive_field(self, patch, resort_mask, upscale_size):
        f_pic = np.zeros((upscale_size)).flatten()
        f_pic[resort_mask] = patch.flatten()
        return f_pic.reshape(self.get_pattern_dimension())

    def fill_gaps_simple(self, image):
        from scipy import signal
        decomp_img = image / np.max(image)
        blured = signal.convolve2d(image, gkern(int(np.max(image.shape) / 10), np.maximum(np.max(image.shape) / 50, 1)),boundary='symm', mode='same')
        blured = blured / np.max(blured)  # * 1.3
        return decomp_img + blured * (decomp_img == 0)

    def initialize(self):
        super().initialize()

        self.foveation_upscale = self.kwargs.get('foveation_upscale', 1.0)
        self.foveation_sharpness = self.kwargs.get('foveation_sharpness', 1.5)
        if self.foveation_upscale > 1.0:
            self.wup = int(self.grid_width * self.foveation_upscale)
            self.hup = int(self.grid_height * self.foveation_upscale)
            self.retina_pic, self.foveation_mask = self.get_receptive_field_resort_mask(self.wup, self.hup, self.grid_width, self.grid_height, self.foveation_sharpness, self.grid_channels)
        else:
            self.foveation_mask = True
            self.wup = self.grid_width
            self.hup = self.grid_height

    def get_pattern_values(self, pattern):
        x = np.random.randint(pattern.shape[2] - self.wup)
        y = np.random.randint(pattern.shape[1] - self.hup)
        #import matplotlib.pyplot as plt
        #plt.matshow(pattern[:, y:y + self.hup, x:x + self.wup].reshape((450*2,450)), cmap=plt.cm.gist_gray)
        return self.compress_receptive_field(pattern[:, y:y + self.hup, x:x + self.wup], self.foveation_mask)

    def get_retina_pic_dimension(self):
        return (self.wup, self.hup)

    def get_retina_pic(self):
        result=np.zeros(self.get_pattern_dimension()).flatten()
        if self.foveation_upscale > 1.0:
            result[self.foveation_mask]+=1
        return result.reshape(self.get_vstackpattern_dimension())

    def get_pattern_dimension(self):
        return (self.grid_channels, self.hup, self.wup)

    def reconstruct_pattern(self, p):
        return self.decompress_receptive_field(p, self.foveation_mask, self.wup * self.hup * self.grid_channels).reshape(self.get_pattern_dimension())



import sys
sys.path.append('../../')

from PymoNNto.Exploration.Evolution_Old.Evolution import *
from multiprocessing import Process, Queue, Pipe
import PymoNNto.Exploration.Evolution_Old.Computing_Devices as comp_dev
import subprocess
import time
import os


def multithreading_evaluation_func_wrapper(param, conn):
    #try:
    #    import psutil as psu
    #    parent = psu.Process()
    #    parent.nice(psu.BELOW_NORMAL_PRIORITY_CLASS)
    #except:
    #    print('not able to set BELOW_NORMAL_PRIORITY_CLASS')

    exec('import ' + param.get('import_file') + ' as target')
    eval_f = eval('target.'+param.get('func_name'))

    while True:

        if conn.poll():
            individual=conn.recv()
            try:
                param['ind']=individual
                print(param)
                fitness = eval_f(attrs=param)

                conn.send([fitness, individual])
            except:
                time.sleep(10.0)
                conn.send([None, individual])

        conn.send('ready')
        time.sleep(1.0)



def remote_evaluation_func_wrapper(device, param, conn):

    while True:
        if conn.poll():
            individual = conn.recv()
            try:

                name=param.get('name', None)

                pexec = 'python3 exec.py'
                for k, v in param.items():#add parameters to command
                    pexec += ' '+k+'='+str(v).replace(' ', '')
                pexec += ' ind=' + str(individual).replace(' ', '')  # add current individual

                if hasattr(device, 'slurm_wrapper'):
                    pexec = device.get_slurm_Evo_Execute_cmd(name, pexec, command='srun')

                output = device.exec_cmd('cd '+device.main_path+name+'/Exploration/Evolution_Old/; '+pexec)

                fitness = output.split(' ')[-1].replace('\r', '').replace('\n', '')
                if fitness.replace('.', '').isnumeric():
                    fitness = float(fitness)
                else:
                    fitness = None
                conn.send([fitness, individual])

            except:
                time.sleep(10.0)
                conn.send([None, individual])


        conn.send('ready')
        time.sleep(1.0)



def run_multiple_times(function, num_threads, args=[]):
    if num_threads == -1:
        function(args)#todo args???
    else:
        result = []
        for i in range(num_threads):
            p = Process(target=function, args=args)
            p.start()
            result.append(p)
        return p

class Multithreaded_Evolution(Evolution):

    def add_process(self):
        parent_conn, child_conn = Pipe()
        p = Process(target=multithreading_evaluation_func_wrapper, args=(self.param, child_conn))
        p.pipe_conn = parent_conn
        p.start()
        self.rendering_threads.append(p)

    def add_remote_process(self, device):
        parent_conn, child_conn = Pipe()
        p = Process(target=remote_evaluation_func_wrapper, args=(device, self.param, child_conn))
        p.pipe_conn = parent_conn
        p.start()
        self.remote_rendering_threads.append(p)

    #def remove_process(self):
    #    p = self.rendering_threads.pop()
    #    p.terminate()
    #    self.individual_pipelines.pop()
    #    self.fitness_pipelines.pop()

    #def adjust_process_count(self):
    #    for _ in range(len(self.rendering_threads) - self.thread_count):
    #        self.add_process()#

    #    for _ in range(self.thread_count - len(self.rendering_threads)):
    #        self.remove_process()


    def __init__(self, import_file, func_name, max_individual_count, generations=None, thread_count=1, name='Evolution_Old', mutation=0.1, constraints=[], death_rate=0.5, param={}, distributed=False):
        #exec('from ' + import_file + ' import *')

        self.import_file = import_file
        self.func_name = func_name

        #ensure that both parameters exist in param file
        param['import_file'] = import_file
        param['func_name'] = func_name

        super().__init__(None, max_individual_count, generations, name, mutation=mutation, constraints=constraints, death_rate=death_rate, param=param)

        self.rendering_threads = []
        #self.individual_pipelines = []
        #self.fitness_pipelines = []

        for t_i in range(thread_count):
            self.add_process()

        self.remote_rendering_threads = []

        if distributed:
            for dev in comp_dev.get_devices(local=False):
                for _ in range(dev.cores):
                    self.add_remote_process(dev)

    def get_thread_state(self, thread):
        ready = False
        data = None
        while thread.pipe_conn.poll():
            read = thread.pipe_conn.recv()#0.444988766308
            if read == 'ready':
                ready = True
            else:
                data = read
        return ready, data

    def get_fitnesses(self, living_individuals):
        #print()
        last_progress=-1

        #self.adjust_process_count()

        fitnesses = [None for _ in living_individuals]

        while None in fitnesses or -1 in fitnesses or -2 in fitnesses:

            for thread in self.rendering_threads:

                ready, data = self.get_thread_state(thread)

                if ready:
                    if None in fitnesses:
                        index = fitnesses.index(None)
                        fitnesses[index] = -1
                        individual = living_individuals[index]
                        thread.pipe_conn.send(individual)
                    elif -2 in fitnesses:
                        index = fitnesses.index(-2)
                        fitnesses[index] = -1
                        individual = living_individuals[index]
                        thread.pipe_conn.send(individual)

                if data is not None:
                    fitness, individual = data
                    if individual in living_individuals:
                        index = living_individuals.index(individual)
                        fitnesses[index] = fitness
                        #print(fitness, individual)

                time.sleep(0.1)

            for thread in self.remote_rendering_threads:

                ready, data = self.get_thread_state(thread)

                if ready:
                    if None in fitnesses:
                        index = fitnesses.index(None)
                        fitnesses[index] = -2
                        individual = living_individuals[index]
                        thread.pipe_conn.send(individual)

                if data is not None:
                    fitness, individual = data
                    if individual in living_individuals:
                        index = living_individuals.index(individual)
                        fitnesses[index] = fitness
                        # print(fitness, individual)

                time.sleep(0.1)

            current_progress = sum([1 for i in fitnesses if i is not None and i>=0])/len(fitnesses)
            if current_progress != last_progress:
                print('\rprogress: '+str(current_progress)+'%', end='')
                last_progress = current_progress

            #print(fitnesses)
            time.sleep(1.0)  # 0.1

        return fitnesses


def args_to_param_dict(add=[]):
    result = {}
    for arg in add+sys.argv[1:]:
        split = arg.split('=')
        if len(split) == 2:
            key, value  =split
            try:
                result[key] = eval(value)
            except:
                result[key] = value
        else:
            print('unknown parameter', arg)
    return result




if __name__ == '__main__':

    params = args_to_param_dict(['name=default_evo_name', 'evolution=True', 'N_e=1400', 'TS=[1]', 'print=False', 'UI=False', 'ff=True', 'fb=True', 'plastic=30000'])

    name = params.get('name', None)
    import_file = params.get('import_file', 'Testing.SORN_Grammar.KWTA_SORN_Cluster')#GrammarExperiment_Hierarchical
    mutation = params.get('mutation', 0.05)
    distributed = params.get('distributed', False)
    max_individual_count = params.get('max_individual_count', 30)
    thread_count = params.get('thread_count', 5)
    func_name = params.get('func_name', 'run')
    individuals = params.get('individuals', [[0.1, 0.02, 0.2944, 0.007, 0.00015, 0.00015, 0.3]])#[[0.95, 0.4, 0.1383, 0.1698, 0.1, 0.00015, 0.04, 0.2944, 0.0006, 0.5, 0.015, 0.2944, 0.1, 0.0001, 0.87038, 0.82099, 1.5, 1.5, 0.08, 15.0],[0.870120065962941, 0.4667924963782921, 0.14228607543158428, 0.14649015102852803, 0.06949631601080168, 0.00012877439537789122, 0.05120564376660136, 0.30655861883836777, 0.00048159856597856514, 0.5134970163375102, 0.014700013181682662, 0.24567552614678428, 0.07771714822925117, 9.899704117749138e-05, 1.2869594699456024, 1.0192332336953436, 3.1214784390540715, 1.3681795404840476, 0.07977142902067157, 17.01811702309001],[0.95, 0.4, 0.1383, 0.1698, 0.1, 0.00015, 0.04, 0.2944, 0.0006, 0.5, 0.015, 0.2944, 0.1, 0.0001, 0.87038, 0.82099, 0.1838, 1.5, 0.08, 15.0],[0.9017392286400007, 0.336521535781792, 0.13270199949220565, 0.14962388907357316, 0.06263567358824208, 0.0001397543247035365, 0.05482554055188729, 0.18435824905204146, 0.00042698042629804394, 0.3583295613950753, 0.01496552477298455, 0.25978647637282687, 0.09257143917052511, 0.0001021643241400783, 1.2256750667338105, 0.9819249648687961, 2.7626521922657132, 2.7626521922657132, 0.08020358998634196, 12.47920274827478]]
    if type(individuals[0]) is not list:
        individuals = [individuals]

    evolution = Multithreaded_Evolution(import_file=import_file, func_name=func_name, max_individual_count=max_individual_count, thread_count=thread_count, name=name, mutation=mutation, constraints=['ind=np.clip(ind,0.00001,None)'], distributed=distributed, param=params)

    if evolution.evo_file_exists():
        evolution.continue_evo(name+'.txt')
    else:
        evolution.start(individuals)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="PymoNNto",
    version="0.9.4",
    author="Marius Vieth",
    author_email="mv15go@gmail.com",
    description="Python Modular Neural Network Toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/trieschlab/PymoNNto",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'PyQt5', 'pyqtgraph>=0.11.0rc0', 'matplotlib', 'scipy', 'scikit-learn', 'imageio', 'pillow'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from PymoNNto.Exploration.Network_UI.TabBase import *

class multi_group_plot_tab(TabBase):

    def __init__(self, variables, line_dict={}, title='Multi Group', timesteps=500):
        super().__init__(title)
        self.variables = variables
        self.line_dict = line_dict
        self.timesteps = timesteps

        #if tensorflow:
        #    self.add_str='.numpy()'
        #else:
        #    self.add_str=''

    def add_recorder_variables(self, neuron_group, Network_UI):
        for var in self.variables:
            try:
                print(var)
                Network_UI.add_recording_variable(neuron_group, 'n.'+var, timesteps=self.timesteps)
                Network_UI.add_recording_variable(neuron_group, 'np.mean(n.'+var+')', timesteps=self.timesteps)
            except:
                print('cannot add', var)



    def initialize(self, Network_UI):
        self.main_tab = Network_UI.Next_Tab(self.title)

        #self.sliders = []
        #lines = []



        #if len(Network_UI.network['IPTI'])>0:

        self.net_var_curves={}
        for i, var in enumerate(self.variables):
            stretch=1
            if i==0:
                stretch=2

            if var in self.line_dict:
                ld_val=self.line_dict[var]
                if type(ld_val) is list:
                    lines=ld_val
                else:
                    lines=[ld_val]
            else:
                lines=[]

            self.net_var_curves[var] = Network_UI.Add_plot_curve(stretch=stretch, number_of_curves=len(Network_UI.neuron_visible_groups), return_list=True, x_label='t (iterations)', y_label='Network average ' + var, lines=lines)


        Network_UI.Next_H_Block()

        self.neuron_var_curves = {}
        for i, var in enumerate(self.variables):
            stretch = 1
            if i == 0:
                stretch = 2

            if var in self.line_dict:
                ld_val = self.line_dict[var]
                if type(ld_val) is list:
                    lines = ld_val
                else:
                    lines = [ld_val]
            else:
                lines = []

            self.neuron_var_curves[var] = Network_UI.Add_plot_curve(stretch=stretch, colors=[Network_UI.neuron_select_color], legend=False, x_label='t (iterations)', y_label='Neuron ' + var, lines=lines)



    def update(self, Network_UI):
        if self.main_tab.isVisible():


            for i, group_tag in enumerate(Network_UI.neuron_visible_groups):
                if len(Network_UI.network[group_tag]) > 0:
                    group=Network_UI.network[group_tag, 0]
                    squeeze= Network_UI.group_sliders[i].sliderPosition() / 100
                    #squeeze = self.sliders[i].sliderPosition() / 100

                    for var in self.variables:
                        try:#if hasattr(group, var):
                            net_data = group['np.mean(n.'+var+')', 0, 'np'][-self.timesteps:]
                            iterations = group['n.iteration', 0, 'np'][-self.timesteps:]
                            self.net_var_curves[var][i].setData(iterations, net_data * squeeze, pen=group.color)
                        except:#else:
                            self.net_var_curves[var][i].clear()


            group = Network_UI.network[Network_UI.neuron_select_group, 0]

            for var in self.variables:
                try:#if hasattr(group, var):
                    neuron_data = group['n.' + var, 0, 'np'][-self.timesteps:]
                    iterations = group['n.iteration', 0, 'np'][-self.timesteps:]
                    if len(neuron_data.shape) > 1:
                        neuron_data = neuron_data[:, Network_UI.neuron_select_id]
                    self.neuron_var_curves[var].setData(iterations, neuron_data)
                except:  #else:
                    self.neuron_var_curves[var].clear()




from PymoNNto.NetworkCore.Behaviour import *
from PymoNNto.NetworkBehaviour.Logic.EulerEquationModules.Helper.Helper import *
import numpy as np
from sympy import symbols
import sympy.physics.units as units
from sympy import symbols
from sympy.physics.units import *

class EquationModule(Behaviour):

    def set_variables(self, neurons):
        n=neurons
        self.add_tag('EquationModule')
        self.step_size = self.get_init_attr('step_size', '1*ms', neurons)
        eq_parts = eq_split(self.get_init_attr('eq', None))

        if eq_parts[0][0] == 'd' and eq_parts[1] == '/' and eq_parts[2] == 'dt' and eq_parts[3] == '=':
            self.variable = eq_parts[0][1:]
            symbols('n.' + self.variable)
        else:
            print('invalid equation')

        for i in range(4, len(eq_parts)):
            if eq_parts[i] in neurons.__dict__:
                eq_parts[i] = 'n.'+eq_parts[i]

        eq_parts = remove_units(eq_parts, 4)


        self.evaluation = 'n.' + self.variable + '+(' + ''.join(eq_parts[4:])+')*{}'.format(neurons.clock_step_size)

        self.compiled_evaluation=compile(self.evaluation, '<string>', 'eval')

        #self.evaluation += ''#*{}'.format(neurons['Clock', 0].)
        print(self.evaluation)
        #print('conv:',convert_to(eval(self.evaluation),second))


    def new_iteration(self, n):
        new = eval(self.compiled_evaluation)
        setattr(n, self.variable+'_new', new)


